# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
All the `deadline job` commands.
"""

from __future__ import annotations
import json
import logging
from configparser import ConfigParser
from pathlib import Path
import os
import sys
from typing import Callable, Optional, Union
import datetime
from typing import Any

import click
from botocore.exceptions import ClientError

from deadline.client.api._session import _modified_logging_level
from deadline.job_attachments.download import OutputDownloader
from deadline.job_attachments.models import (
    FileConflictResolution,
    JobAttachmentS3Settings,
    PathFormat,
)
from deadline.job_attachments.progress_tracker import (
    DownloadSummaryStatistics,
    ProgressReportMetadata,
)
from deadline.job_attachments._utils import _human_readable_file_size

from ... import api
from ...config import config_file
from ...exceptions import DeadlineOperationError
from .._common import _apply_cli_options_to_config, _cli_object_repr, _handle_error
from ._sigint_handler import SigIntHandler

JSON_MSG_TYPE_TITLE = "title"
JSON_MSG_TYPE_PRESUMMARY = "presummary"
JSON_MSG_TYPE_PATH = "path"
JSON_MSG_TYPE_PATHCONFIRM = "pathconfirm"
JSON_MSG_TYPE_PROGRESS = "progress"
JSON_MSG_TYPE_SUMMARY = "summary"
JSON_MSG_TYPE_ERROR = "error"


# Set up the signal handler for handling Ctrl + C interruptions.
sigint_handler = SigIntHandler()


@click.group(name="job")
@_handle_error
def cli_job():
    """
    Commands to work with jobs.
    """


@cli_job.command(name="list")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--page-size", default=5, help="The number of jobs to load at a time.")
@click.option("--item-offset", default=0, help="The index of the job to start listing from.")
@_handle_error
def job_list(page_size, item_offset, **args):
    """
    Lists the Jobs in a queue.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(required_options={"farm_id", "queue_id"}, **args)

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    try:
        response = deadline.search_jobs(
            farmId=farm_id,
            queueIds=[queue_id],
            itemOffset=item_offset,
            pageSize=page_size,
            sortExpressions=[{"fieldSort": {"name": "CREATED_AT", "sortOrder": "DESCENDING"}}],
        )
    except ClientError as exc:
        raise DeadlineOperationError(f"Failed to get Jobs from Deadline:\n{exc}") from exc

    total_results = response["totalResults"]

    # Select which fields to print and in which order
    name_field = "displayName"
    if len(response["jobs"]) and "name" in response["jobs"][0]:
        name_field = "name"
    structured_job_list = [
        {
            field: job.get(field, "")
            for field in [
                name_field,
                "jobId",
                "taskRunStatus",
                "startedAt",
                "endedAt",
                "createdBy",
                "createdAt",
            ]
        }
        for job in response["jobs"]
    ]

    click.echo(
        f"Displaying {len(structured_job_list)} of {total_results} Jobs starting at {item_offset}"
    )
    click.echo()
    click.echo(_cli_object_repr(structured_job_list))


@cli_job.command(name="get")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to get.")
@_handle_error
def job_get(**args):
    """
    Get the details of a job.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    deadline = api.get_boto3_client("deadline", config=config)
    response = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    response.pop("ResponseMetadata", None)

    click.echo(_cli_object_repr(response))


@cli_job.command(name="cancel")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to cancel.")
@click.option(
    "--mark-as",
    type=click.Choice(["CANCELED", "FAILED", "SUCCEEDED"], case_sensitive=False),
    default="CANCELED",
    help="The status to apply to all active tasks in the job.",
)
@click.option(
    "--yes",
    is_flag=True,
    help="Skip any confirmation prompts",
)
@_handle_error
def job_cancel(mark_as: str, yes: bool, **args):
    """
    Cancel job from running.
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    mark_as = mark_as.upper()

    deadline = api.get_boto3_client("deadline", config=config)

    # Print a summary of the job to cancel
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    # Remove the zero-count status counts
    job["taskRunStatusCounts"] = {
        name: count for name, count in job["taskRunStatusCounts"].items() if count != 0
    }
    # Filter the fields to a summary
    filtered_job = {
        field: job.get(field, "")
        for field in [
            "name",
            "jobId",
            "taskRunStatus",
            "taskRunStatusCounts",
            "startedAt",
            "endedAt",
            "createdBy",
            "createdAt",
        ]
    }
    click.echo(_cli_object_repr(filtered_job))

    # Ask for confirmation about canceling this job.
    if not (
        yes or config_file.str2bool(config_file.get_setting("settings.auto_accept", config=config))
    ):
        if mark_as == "CANCELED":
            cancel_message = "Are you sure you want to cancel this job?"
        else:
            cancel_message = (
                f"Are you sure you want to cancel this job and mark its taskRunStatus as {mark_as}?"
            )
        # We explicitly require a yes/no response, as this is an operation that will interrupt the work in progress
        # on their job.
        if not click.confirm(
            cancel_message,
            default=None,
        ):
            click.echo("Job not canceled.")
            sys.exit(1)

    if mark_as == "CANCELED":
        click.echo("Canceling job...")
    else:
        click.echo(f"Canceling job and marking as {mark_as}...")
    deadline.update_job(farmId=farm_id, queueId=queue_id, jobId=job_id, targetTaskRunStatus=mark_as)


def _download_job_output(
    config: Optional[ConfigParser],
    farm_id: str,
    queue_id: str,
    job_id: str,
    step_id: Optional[str],
    task_id: Optional[str],
    is_json_format: bool = False,
):
    """
    Starts the download of job output and handles the progress reporting callback.
    """
    deadline = api.get_boto3_client("deadline", config=config)

    auto_accept = config_file.str2bool(
        config_file.get_setting("settings.auto_accept", config=config)
    )
    conflict_resolution = config_file.get_setting("settings.conflict_resolution", config=config)

    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    step = {}
    task = {}
    if step_id:
        step = deadline.get_step(farmId=farm_id, queueId=queue_id, jobId=job_id, stepId=step_id)
    if task_id:
        task = deadline.get_task(
            farmId=farm_id, queueId=queue_id, jobId=job_id, stepId=step_id, taskId=task_id
        )

    click.echo(
        _get_start_message(job["name"], step.get("name"), task.get("parameters"), is_json_format)
    )

    queue = deadline.get_queue(farmId=farm_id, queueId=queue_id)

    queue_role_session = api.get_queue_user_boto3_session(
        deadline=deadline,
        config=config,
        farm_id=farm_id,
        queue_id=queue_id,
        queue_display_name=queue["displayName"],
    )

    # Get a dictionary mapping rootPath to rootPathFormat (OS) from job's manifests
    root_path_format_mapping: dict[str, str] = {}
    job_attachments = job.get("attachments", None)
    if job_attachments:
        job_attachments_manifests = job_attachments["manifests"]
        for manifest in job_attachments_manifests:
            root_path_format_mapping[manifest["rootPath"]] = manifest["rootPathFormat"]

    job_output_downloader = OutputDownloader(
        s3_settings=JobAttachmentS3Settings(**queue["jobAttachmentSettings"]),
        farm_id=farm_id,
        queue_id=queue_id,
        job_id=job_id,
        step_id=step_id,
        task_id=task_id,
        session=queue_role_session,
    )

    output_paths_by_root = job_output_downloader.get_output_paths_by_root()

    # If no output paths were found, log a message and exit.
    if output_paths_by_root == {}:
        click.echo(_get_no_output_message(is_json_format))
        return

    # Check if the asset roots came from different OS. If so, prompt users to
    # select alternative root paths to download to, (regardless of the auto-accept.)
    asset_roots = list(output_paths_by_root.keys())
    for asset_root in asset_roots:
        root_path_format = root_path_format_mapping.get(asset_root, "")
        if root_path_format == "":
            # There must be a corresponding root path format for each root path, by design.
            raise DeadlineOperationError(f"No root path format found for {asset_root}.")
        if PathFormat.get_host_path_format_string() != root_path_format:
            click.echo(_get_mismatch_os_root_warning(asset_root, root_path_format, is_json_format))

            if not is_json_format:
                new_root = click.prompt(
                    "> Please enter a new root path",
                    type=click.Path(exists=False),
                )
            else:
                json_string = click.prompt("", prompt_suffix="", type=str)
                new_root = _get_value_from_json_line(
                    json_string, JSON_MSG_TYPE_PATHCONFIRM, expected_size=1
                )[0]
                _assert_valid_path(new_root)

            job_output_downloader.set_root_path(asset_root, new_root)

    output_paths_by_root = job_output_downloader.get_output_paths_by_root()

    # Prompt users to confirm local root paths where they will download outputs to,
    # and allow users to select different location to download files to if they want.
    # (If auto-accept is enabled, automatically download to the default root paths.)
    if not auto_accept:
        if not is_json_format:
            user_choice = ""
            while user_choice != ("y" or "n"):
                click.echo(
                    _get_summary_of_files_to_download_message(output_paths_by_root, is_json_format)
                )
                asset_roots = list(output_paths_by_root.keys())
                click.echo(_get_roots_list_message(asset_roots, is_json_format))
                user_choice = click.prompt(
                    "> Please enter the index of root directory to edit, y to proceed without changes, or n to cancel the download",
                    type=click.Choice(
                        [*[str(num) for num in list(range(0, len(asset_roots)))], "y", "n"]
                    ),
                    default="y",
                )
                if user_choice == "n":
                    click.echo("Output download canceled.")
                    return
                elif user_choice != "y":
                    # User selected an index to modify the root directory.
                    index_to_change = int(user_choice)
                    new_root = click.prompt(
                        "> Please enter the new root directory path, or press Enter to keep it unchanged",
                        type=click.Path(exists=False),
                        default=asset_roots[index_to_change],
                    )
                    job_output_downloader.set_root_path(
                        asset_roots[index_to_change], str(Path(new_root))
                    )
                    output_paths_by_root = job_output_downloader.get_output_paths_by_root()
        else:
            click.echo(
                _get_summary_of_files_to_download_message(output_paths_by_root, is_json_format)
            )
            asset_roots = list(output_paths_by_root.keys())
            click.echo(_get_roots_list_message(asset_roots, is_json_format))
            json_string = click.prompt("", prompt_suffix="", type=str)
            confirmed_asset_roots = _get_value_from_json_line(
                json_string, JSON_MSG_TYPE_PATHCONFIRM, expected_size=len(asset_roots)
            )
            for index, confirmed_root in enumerate(confirmed_asset_roots):
                _assert_valid_path(confirmed_root)
                job_output_downloader.set_root_path(asset_roots[index], str(Path(confirmed_root)))
            output_paths_by_root = job_output_downloader.get_output_paths_by_root()

    # If the conflict resolution option was not specified, auto-accept is false, and
    # if there are any conflicting files in local, prompt users to select a resolution method.
    # (skip, overwrite, or make a copy.)
    if conflict_resolution != FileConflictResolution.NOT_SELECTED.name:
        file_conflict_resolution = FileConflictResolution[conflict_resolution]
    elif auto_accept:
        file_conflict_resolution = FileConflictResolution.CREATE_COPY
    else:
        file_conflict_resolution = FileConflictResolution.CREATE_COPY
        conflicting_filenames = _get_conflicting_filenames(output_paths_by_root)
        if conflicting_filenames:
            click.echo(_get_conflict_resolution_selection_message(conflicting_filenames))
            user_choice = click.prompt(
                "> Please enter your choice (1, 2, 3, or n to cancel the download)",
                type=click.Choice(["1", "2", "3", "n"]),
                default="3",
            )
            if user_choice == "n":
                click.echo("Output download canceled.")
                return
            else:
                resolution_choice_int = int(user_choice)
                file_conflict_resolution = FileConflictResolution(resolution_choice_int)

    # TODO: remove logging level setting when the max number connections for boto3 client
    # in Job Attachments library can be increased (currently using default number, 10, which
    # makes it keep logging urllib3 warning messages when downloading large files)
    with _modified_logging_level(logging.getLogger("urllib3"), logging.ERROR):

        @api.record_success_fail_telemetry_event(metric_name="download_job_output")  # type: ignore
        def _download_job_output(
            file_conflict_resolution: Optional[
                FileConflictResolution
            ] = FileConflictResolution.CREATE_COPY,
            on_downloading_files: Optional[Callable[[ProgressReportMetadata], bool]] = None,
        ) -> DownloadSummaryStatistics:
            return job_output_downloader.download_job_output(
                file_conflict_resolution=file_conflict_resolution,
                on_downloading_files=on_downloading_files,
            )

        if not is_json_format:
            # Note: click doesn't export the return type of progressbar(), so we suppress mypy warnings for
            # not annotating the type of download_progress.
            click.echo()
            with click.progressbar(length=100, label="Downloading Outputs") as download_progress:  # type: ignore[var-annotated]

                def _update_download_progress(download_metadata: ProgressReportMetadata) -> bool:
                    new_progress = int(download_metadata.progress) - download_progress.pos
                    if new_progress > 0:
                        download_progress.update(new_progress)
                    return sigint_handler.continue_operation

                download_summary: DownloadSummaryStatistics = _download_job_output(  # type: ignore
                    file_conflict_resolution=file_conflict_resolution,
                    on_downloading_files=_update_download_progress,
                )
        else:

            def _update_download_progress(download_metadata: ProgressReportMetadata) -> bool:
                click.echo(
                    _get_json_line(JSON_MSG_TYPE_PROGRESS, str(int(download_metadata.progress)))
                )
                # TODO: enable download cancellation for JSON format
                return True

            download_summary = _download_job_output(  # type: ignore
                file_conflict_resolution=file_conflict_resolution,
                on_downloading_files=_update_download_progress,
            )

    click.echo(_get_download_summary_message(download_summary, is_json_format))
    click.echo()


def _get_start_message(
    job_name: str, step_name: Optional[str], task_parameters: Optional[dict], is_json_format: bool
) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_TITLE, job_name)
    else:
        if step_name is None:
            return f"Downloading output from Job {job_name!r}"
        elif task_parameters is None:
            return f"Downloading output from Job {job_name!r} Step {step_name!r}"
        else:
            task_parameters_summary = "{}"
            if task_parameters:
                task_parameters_summary = (
                    "{"
                    + ",".join(
                        f"{key}={list(value.values())[0]}" for key, value in task_parameters.items()
                    )
                    + "}"
                )
            return f"Downloading output from Job {job_name!r} Step {step_name!r} Task {task_parameters_summary}"


def _get_no_output_message(is_json_format: bool) -> str:
    msg = (
        "There are no output files available for download at this moment. Please verify that"
        " the Job/Step/Task you are trying to download output from has completed successfully."
    )
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_SUMMARY, msg)
    else:
        return msg


def _get_mismatch_os_root_warning(root: str, root_path_format: str, is_json_format: bool) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PATH, [root])
    else:
        path_format_capitalized_first_letter = root_path_format[0].upper() + root_path_format[1:]
        return (
            "This root path format does not match the operating system you're using. "
            "Where would you like to save the files?\n"
            f"The location was {root}, on {path_format_capitalized_first_letter}."
        )


def _get_summary_of_files_to_download_message(
    output_paths_by_root: dict[str, list[str]], is_json_format: bool
) -> str:
    # Print some information about what we will download
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PRESUMMARY, output_paths_by_root)
    else:
        paths_message_joined = "    " + "\n    ".join(
            f"{os.path.commonpath([os.path.join(directory, p) for p in output_paths])} ({len(output_paths)} file{'s' if len(output_paths) > 1 else ''})"
            for directory, output_paths in output_paths_by_root.items()
        )
        return "\n" "Summary of files to download:\n" f"{paths_message_joined}" "\n"


def _get_roots_list_message(asset_roots: list[str], is_json_format: bool) -> str:
    if is_json_format:
        return _get_json_line(JSON_MSG_TYPE_PATH, asset_roots)
    else:
        asset_roots_str = "\n".join([f"[{index}] {root}" for index, root in enumerate(asset_roots)])
        return (
            f"You are about to download files which may come from multiple root directories. Here are a list of the current root directories:\n"
            f"{asset_roots_str}"
        )


def _get_conflict_resolution_selection_message(conflicting_filenames: list[str]) -> str:
    conflicting_filenames_str = "\n        ".join(conflicting_filenames)
    return (
        f"The following files already exist in your local directory:\n        {conflicting_filenames_str}\n"
        f"You have three options to choose from:\n"
        f"[1] Skip: Do not download these files\n"
        f"[2] Overwrite: Download these files and overwrite existing files\n"
        f"[3] Create a copy: Download the file with a new name, appending '(1)' to the end"
    )


def _get_download_summary_message(
    download_summary: DownloadSummaryStatistics, is_json_format: bool
) -> str:
    if is_json_format:
        return _get_json_line(
            JSON_MSG_TYPE_SUMMARY, f"Downloaded {download_summary.processed_files} files"
        )
    else:
        paths_joined = "\n        ".join(
            f"{directory} ({count} file{'s' if count > 1 else ''})"
            for directory, count in download_summary.file_counts_by_root_directory.items()
        )
        return (
            "Download Summary:\n"
            f"    Downloaded {download_summary.processed_files} files totaling"
            f" {_human_readable_file_size(download_summary.processed_bytes)}.\n"
            f"    Total download time of {round(download_summary.total_time, ndigits=5)} seconds"
            f" at {_human_readable_file_size(int(download_summary.transfer_rate))}/s.\n"
            f"    Download locations (total file counts):\n        {paths_joined}"
        )


def _get_conflicting_filenames(filenames_by_root: dict[str, list[str]]) -> list[str]:
    conflicting_filenames: list[str] = []

    for root, filenames in filenames_by_root.items():
        for filename in filenames:
            abs_path = Path(root).joinpath(filename).resolve()
            if abs_path.is_file():
                conflicting_filenames.append(str(abs_path))

    return conflicting_filenames


def _get_json_line(messageType: str, value: Union[str, list[str], dict[str, Any]]) -> str:
    return json.dumps({"messageType": messageType, "value": value}, ensure_ascii=True)


def _get_value_from_json_line(
    json_line: str, message_type: str, expected_size: Optional[int] = None
) -> Union[str, list[str]]:
    try:
        parsed_json = json.loads(json_line)
        if parsed_json["messageType"] != message_type:
            raise ValueError(
                f"Expected message type '{message_type}' but received '{parsed_json['messageType']}'"
            )
        if expected_size and len(parsed_json["value"]) != expected_size:
            raise ValueError(
                f"Expected {expected_size} item{'' if expected_size == 1 else 's'} in value "
                f"but received {len(parsed_json['value'])}"
            )
        return parsed_json["value"]
    except Exception as e:
        raise ValueError(f"Invalid JSON line '{json_line}': {e}")


def _assert_valid_path(path: str) -> None:
    """
    Validates that the path has the format of the OS currently running.
    """
    path_obj = Path(path)
    if not path_obj.is_absolute():
        raise ValueError(f"Path {path} is not an absolute path.")


@cli_job.command(name="download-output")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to use.")
@click.option("--step-id", help="The step to use.")
@click.option("--task-id", help="The task to use.")
@click.option(
    "--conflict-resolution",
    type=click.Choice(
        [
            FileConflictResolution.SKIP.name,
            FileConflictResolution.OVERWRITE.name,
            FileConflictResolution.CREATE_COPY.name,
        ],
        case_sensitive=False,
    ),
    help="How to handle downloads if a file already exists:\n"
    "CREATE_COPY (default): Download the file with a new name, appending '(1)' to the end\n"
    "SKIP: Do not download the file\n"
    "OVERWRITE: Download and replace the existing file",
)
@click.option(
    "--yes",
    is_flag=True,
    help="Skip any confirmation prompts",
)
@click.option(
    "--output",
    type=click.Choice(
        ["verbose", "json"],
        case_sensitive=False,
    ),
    help="Specifies the output format of the messages printed to stdout.\n"
    "VERBOSE: Displays messages in a human-readable text format.\n"
    "JSON: Displays messages in JSON line format, so that the info can be easily "
    "parsed/consumed by custom scripts.",
)
@_handle_error
def job_download_output(step_id, task_id, output, **args):
    """
    Download a job's output.
    """
    if task_id and not step_id:
        raise click.UsageError("Missing option '--step-id' required with '--task-id'")
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)
    is_json_format = True if output == "json" else False

    try:
        _download_job_output(config, farm_id, queue_id, job_id, step_id, task_id, is_json_format)
    except Exception as e:
        if is_json_format:
            error_one_liner = str(e).replace("\n", ". ")
            click.echo(_get_json_line(JSON_MSG_TYPE_ERROR, error_one_liner))
            sys.exit(1)
        else:
            raise DeadlineOperationError(f"Failed to download output:\n{e}") from e


@cli_job.command(name="trace-schedule")
@click.option("--profile", help="The AWS profile to use.")
@click.option("--farm-id", help="The farm to use.")
@click.option("--queue-id", help="The queue to use.")
@click.option("--job-id", help="The job to trace.")
@click.option("-v", "--verbose", is_flag=True, help="Output verbose trace details.")
@click.option(
    "--trace-format",
    type=click.Choice(
        ["chrome"],
        case_sensitive=False,
    ),
    help="The tracing format to write.",
)
@click.option("--trace-file", help="The tracing file to write.")
@_handle_error
def job_trace_schedule(verbose, trace_format, trace_file, **args):
    """
    EXPERIMENTAL - Generate statistics from a completed job.

    To visualize the trace output file when providing the options
    "--trace-format chrome --trace-file <output>.json", use
    the https://ui.perfetto.dev Tracing UI and choose "Open trace file".
    """
    # Get a temporary config object with the standard options handled
    config = _apply_cli_options_to_config(
        required_options={"farm_id", "queue_id", "job_id"}, **args
    )

    farm_id = config_file.get_setting("defaults.farm_id", config=config)
    queue_id = config_file.get_setting("defaults.queue_id", config=config)
    job_id = config_file.get_setting("defaults.job_id", config=config)

    if trace_file and not trace_format:
        raise DeadlineOperationError("Error: Must provide --trace-format with --trace-file.")

    deadline = api.get_boto3_client("deadline", config=config)

    click.echo("Getting the job...")
    job = deadline.get_job(farmId=farm_id, queueId=queue_id, jobId=job_id)
    job.pop("ResponseMetadata", None)

    click.echo("Getting all the sessions for the job...")
    response = deadline.list_sessions(farmId=farm_id, queueId=queue_id, jobId=job_id)
    while "nextToken" in response:
        old_list = response["sessions"]
        response = deadline.list_sessions(
            farmId=farm_id, queueId=queue_id, jobId=job_id, nextToken=response["nextToken"]
        )
        response["sessions"] = old_list + response["sessions"]
    response.pop("ResponseMetadata", None)

    sessions = sorted(response["sessions"], key=lambda session: session["startedAt"])

    click.echo("Getting all the session actions for the job...")
    for session in sessions:
        response = deadline.list_session_actions(
            farmId=farm_id, queueId=queue_id, jobId=job_id, sessionId=session["sessionId"]
        )
        while "nextToken" in response:
            old_list = response["sessionActions"]
            response = deadline.list_session_actions(
                farmId=farm_id,
                queueId=queue_id,
                jobId=job_id,
                sessionId=session["sessionId"],
                nextToken=response["nextToken"],
            )
            response["sessionActions"] = old_list + response["sessionActions"]
        response.pop("ResponseMetadata", None)

        session["actions"] = response["sessionActions"]

    # Cache steps and tasks by their id, to only get each once
    steps: dict[str, Any] = {}
    tasks: dict[str, Any] = {}

    with click.progressbar(  # type: ignore[var-annotated]
        length=len(sessions), label="Getting all the steps and tasks for the job..."
    ) as progressbar:
        for index, session in enumerate(sessions):
            session["index"] = index
            for action in session["actions"]:
                step_id = action["definition"].get("taskRun", {}).get("stepId")
                task_id = action["definition"].get("taskRun", {}).get("taskId")
                if step_id and task_id:
                    if "step" not in session:
                        if step_id in steps:
                            step = steps[step_id]
                        else:
                            step = deadline.get_step(
                                farmId=farm_id, queueId=queue_id, jobId=job_id, stepId=step_id
                            )
                            step.pop("ResponseMetadata", None)
                            steps[step_id] = step
                        session["step"] = step
                    elif session["step"]["stepId"] != step_id:
                        # The session itself doesn't have a step id, but for now the scheduler always creates new
                        # sessions for new steps.
                        raise DeadlineOperationError(
                            f"Session {session['sessionId']} ran more than one step! When this code was"
                            " written that wasn't possible."
                        )

                    if task_id in tasks:
                        task = tasks[task_id]
                    else:
                        task = deadline.get_task(
                            farmId=farm_id,
                            queueId=queue_id,
                            jobId=job_id,
                            stepId=step_id,
                            taskId=task_id,
                        )
                        task.pop("ResponseMetadata", None)
                        tasks[task_id] = task
                    action["task"] = task
            progressbar.update(1)

    # Collect the worker IDs that ran the sessions, and give them indexes to act as PIDs in the tracing file
    worker_ids = {session["workerId"] for session in sessions}
    workers = {worker_id: index for index, worker_id in enumerate(worker_ids)}

    click.echo("Processing the trace data...")
    trace_events = []

    started_at = job["startedAt"]

    def time_int(timestamp: datetime.datetime):
        return int((timestamp - started_at) / datetime.timedelta(microseconds=1))

    def duration_of(resource):
        try:
            return time_int(resource["endedAt"]) - time_int(resource["startedAt"])
        except KeyError:
            return 0

    accumulators = {
        "sessionCount": 0,
        "sessionActionCount": 0,
        "taskRunCount": 0,
        "envActionCount": 0,
        "syncJobAttachmentsCount": 0,
        "sessionDuration": 0,
        "sessionActionDuration": 0,
        "taskRunDuration": 0,
        "envActionDuration": 0,
        "syncJobAttachmentsDuration": 0,
    }

    for session in sessions:
        accumulators["sessionCount"] += 1
        accumulators["sessionDuration"] += duration_of(session)

        pid = workers[session["workerId"]]
        session_event_name = f"{session['step']['name']} - {session['index']}"
        trace_events.append(
            {
                "name": session_event_name,
                "cat": "SESSION",
                "ph": "B",  # Begin Event
                "ts": time_int(session["startedAt"]),
                "pid": pid,
                "tid": 0,
                "args": {
                    "sessionId": session["sessionId"],
                    "workerId": session["workerId"],
                    "fleetId": session["fleetId"],
                    "lifecycleStatus": session["lifecycleStatus"],
                },
            }
        )

        for action in session["actions"]:
            accumulators["sessionActionCount"] += 1
            accumulators["sessionActionDuration"] += duration_of(action)

            name = action["sessionActionId"]
            action_type = list(action["definition"].keys())[0]
            if action_type == "taskRun":
                accumulators["taskRunCount"] += 1
                accumulators["taskRunDuration"] += duration_of(action)

                task = action["task"]
                parameters = task.get("parameters", {})
                name = ",".join(
                    f"{param}={list(parameters[param].values())[0]}" for param in parameters
                )
                if not name:
                    name = "<No Task Params>"
            elif action_type in ("envEnter", "envExit"):
                accumulators["envActionCount"] += 1
                accumulators["envActionDuration"] += duration_of(action)

                name = action["definition"][action_type]["environmentId"].split(":")[-1]
            elif action_type == "syncInputJobAttachments":
                accumulators["syncJobAttachmentsCount"] += 1
                accumulators["syncJobAttachmentsDuration"] += duration_of(action)

                if "stepId" in action["definition"][action_type]:
                    name = "Sync Job Attchmnt (Dependencies)"
                else:
                    name = "Sync Job Attchmnt (Submitted)"
            if "startedAt" in action:
                trace_events.append(
                    {
                        "name": name,
                        "cat": action_type,
                        "ph": "X",  # Complete Event
                        "ts": time_int(action["startedAt"]),
                        "dur": duration_of(action),
                        "pid": pid,
                        "tid": 0,
                        "args": {
                            "sessionActionId": action["sessionActionId"],
                            "status": action["status"],
                            "stepName": session["step"]["name"],
                        },
                    }
                )
        trace_events.append(
            {
                "name": session_event_name,
                "cat": "SESSION",
                "ph": "E",  # End Event
                "ts": time_int(session["endedAt"]),
                "pid": pid,
                "tid": 0,
            }
        )

    if verbose:
        click.echo(" ==== TRACE DATA ====")
        click.echo(_cli_object_repr(job))
        click.echo("")
        click.echo(_cli_object_repr(sessions))

    click.echo("")
    click.echo(" ==== SUMMARY ====")
    click.echo("")
    click.echo(f"Session Count: {accumulators['sessionCount']}")
    session_total_duration = accumulators["sessionDuration"]
    click.echo(f"Session Total Duration: {datetime.timedelta(microseconds=session_total_duration)}")
    click.echo(f"Session Action Count: {accumulators['sessionActionCount']}")
    click.echo(
        f"Session Action Total Duration: {datetime.timedelta(microseconds=accumulators['sessionActionDuration'])}"
    )
    click.echo(f"Task Run Count: {accumulators['taskRunCount']}")
    task_run_total_duration = accumulators["taskRunDuration"]
    click.echo(
        f"Task Run Total Duration: {datetime.timedelta(microseconds=task_run_total_duration)} ({100 * task_run_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(
        f"Non-Task Run Count: {accumulators['sessionActionCount'] - accumulators['taskRunCount']}"
    )
    non_task_run_total_duration = (
        accumulators["sessionActionDuration"] - accumulators["taskRunDuration"]
    )
    click.echo(
        f"Non-Task Run Total Duration: {datetime.timedelta(microseconds=non_task_run_total_duration)} ({100 * non_task_run_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(f"Sync Job Attachments Count: {accumulators['syncJobAttachmentsCount']}")
    sync_job_attachments_total_duration = accumulators["syncJobAttachmentsDuration"]
    click.echo(
        f"Sync Job Attachments Total Duration: {datetime.timedelta(microseconds=sync_job_attachments_total_duration)} ({100 * sync_job_attachments_total_duration / session_total_duration:.1f}%)"
    )
    click.echo(f"Env Action Count: {accumulators['envActionCount']}")
    env_action_total_duration = accumulators["envActionDuration"]
    click.echo(
        f"Env Action Total Duration: {datetime.timedelta(microseconds=env_action_total_duration)} ({100 * env_action_total_duration / session_total_duration:.1f}%)"
    )
    click.echo("")
    within_session_overhead_duration = (
        accumulators["sessionDuration"] - accumulators["sessionActionDuration"]
    )
    click.echo(
        f"Within-session Overhead Duration: {datetime.timedelta(microseconds=within_session_overhead_duration)} ({100 * within_session_overhead_duration / session_total_duration:.1f}%)"
    )
    click.echo(
        f"Within-session Overhead Duration Per Action: {datetime.timedelta(microseconds=(accumulators['sessionDuration'] - accumulators['sessionActionDuration']) / accumulators['sessionActionCount'])}"
    )

    tracing_data: dict[str, Any] = {
        "traceEvents": trace_events,
        # "displayTimeUnits": "s",
        "otherData": {
            "farmId": farm_id,
            "queueId": queue_id,
            "jobId": job_id,
            "jobName": job["name"],
            "startedAt": job["startedAt"].isoformat(sep="T"),
        },
    }
    if "endedAt" in job:
        tracing_data["otherData"]["endedAt"] = job["endedAt"].isoformat(sep="T")

    tracing_data["otherData"].update(accumulators)

    if trace_file:
        with open(trace_file, "w", encoding="utf8") as f:
            json.dump(tracing_data, f, indent=1)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Provides the function to submit a job bundle to AWS Deadline Cloud.
"""
from __future__ import annotations

import json
import logging
import time
import os
import textwrap
from configparser import ConfigParser
from typing import Any, Callable, Dict, List, Optional, Tuple, Union

from botocore.client import BaseClient

from deadline.client.api._job_attachment import _hash_attachments  # type: ignore[import]

from .. import api
from ..exceptions import DeadlineOperationError, CreateJobWaiterCanceled
from ..config import get_setting, set_setting, config_file
from ..job_bundle import deadline_yaml_dump
from ..job_bundle.loader import (
    read_yaml_or_json,
    read_yaml_or_json_object,
    parse_yaml_or_json_content,
    validate_directory_symlink_containment,
)
from ..job_bundle.parameters import (
    apply_job_parameters,
    merge_queue_job_parameters,
    read_job_bundle_parameters,
    JobParameter,
)
from ..job_bundle.submission import AssetReferences, split_parameter_args
from ...job_attachments.exceptions import MisconfiguredInputsError
from ...job_attachments.models import (
    JobAttachmentsFileSystem,
    AssetRootManifest,
    AssetUploadGroup,
    JobAttachmentS3Settings,
)
from ...job_attachments.progress_tracker import ProgressReportMetadata
from ...job_attachments.upload import S3AssetManager
from ._session import session_context

logger = logging.getLogger(__name__)


def create_job_from_job_bundle(
    job_bundle_dir: str,
    job_parameters: list[dict[str, Any]] = [],
    *,
    name: Optional[str] = None,
    queue_parameter_definitions: Optional[list[JobParameter]] = None,
    job_attachments_file_system: Optional[str] = None,
    config: Optional[ConfigParser] = None,
    priority: Optional[int] = None,
    max_failed_tasks_count: Optional[int] = None,
    max_retries_per_task: Optional[int] = None,
    print_function_callback: Callable[[str], None] = lambda msg: None,
    decide_cancel_submission_callback: Callable[
        [AssetUploadGroup], bool
    ] = lambda upload_group: False,
    hashing_progress_callback: Optional[Callable[[ProgressReportMetadata], bool]] = None,
    upload_progress_callback: Optional[Callable[[ProgressReportMetadata], bool]] = None,
    create_job_result_callback: Optional[Callable[[], bool]] = None,
    require_paths_exist: bool = False,
    submitter_name: Optional[str] = None,
) -> Union[str, None]:
    """
    Creates a job in the farm/queue configured as default for the
    workstation from the job bundle in the provided directory.

    A job bundle has the following directory structure:

    /template.json|yaml (required): An Open Job Description job template that specifies the work to be done. Job parameters
            are embedded here.
    /parameter_values.json|yaml (optional): If provided, these are parameter values for the job template and for
            the render farm. AWS Deadline Cloud-specific parameters are like "deadline:priority".
            Looks like:
            {
                "parameterValues": [
                    {"name": "<name>", "value": "<value>"},
                    ...
                ]
            }
    /asset_references.json|yaml (optional): If provided, these are references to the input and output assets
            of the job. Looks like:
            {
                "assetReferences": {
                    "inputs": {
                        "filenames": [
                            "/mnt/path/to/file.txt",
                            ...
                        ],
                        "directories": [
                            "/mnt/path/to/directory",
                            ...
                        ],
                    },
                    "outputs": {
                        "directories": [
                            "/mnt/path/to/output_directory",
                            ...
                        ],
                    }
                }
            }

    Args:
        job_bundle_dir (str): The directory containing the job bundle.
        job_parameters (List[Dict[str, Any]], optional): A list of job parameters in the following format:
            [{"name": "<name>", "value": "<value>"}, ...]
        name (str, optional): The name of the job to submit, replacing the name defined in the job bundle.
        queue_parameter_definitions (list[JobParameter], optional) A list of queue_parameters to use
                instead of retrieving queue_parameters from the queue with get_queue_parameter_definitions.
        job_attachments_file_system (str, optional): define which file system to use;
                (valid values: "COPIED", "VIRTUAL") instead of using the value in the config file.
        config (ConfigParser, optional): The AWS Deadline Cloud configuration
                object to use instead of the config file.
        priority (int, optional): explicit value for the priority of the job.
        max_failed_tasks_count (int, optional): explicit value for the maximum allowed failed tasks.
        max_retries_per_task (int, optional): explicit value for the maximum retries per task.
        print_function_callback (Callable str -> None, optional): Callback to print messages produced in this function.
                Used in the CLI to print to stdout using click.echo. By default ignores messages.
        decide_cancel_submission_callback (Callable dict[str, int], int, int -> bool): If the job has job
                attachments, decide whether or not to cancel the submission given what assets will
                or will not be uploaded. If returns true, the submission is canceled. If False,
                the submission continues. By default the submission always continues.
        hashing_progress_callback / upload_progress_callback / create_job_result_callback (Callable -> bool):
                Callbacks periodically called while hashing / uploading / waiting for job creation. If returns false,
                the operation will be cancelled. If return true, the operation continues. Default behavior for each
                is to not cancel the operation. hashing_progress_callback and upload_progress_callback both receive
                ProgressReport as a parameter, which can be used for projecting remaining time, as in done in the CLI.
    """

    if not submitter_name:
        submitter_name = "CLI"

    session_context["submitter-name"] = submitter_name

    # Ensure the job bundle doesn't contain files that resolve outside of the bundle directory
    validate_directory_symlink_containment(job_bundle_dir)

    # Read in the job template
    file_contents, file_type = read_yaml_or_json(job_bundle_dir, "template", required=True)

    # If requested, substitute the job name in the template
    if name is not None:
        template_obj = parse_yaml_or_json_content(
            file_contents, file_type, job_bundle_dir, "template"
        )
        template_obj["name"] = name
        if file_type == "YAML":
            file_contents = deadline_yaml_dump(template_obj)
        else:
            file_contents = json.dumps(template_obj)

    deadline = api.get_boto3_client("deadline", config=config)
    queue_id = get_setting("defaults.queue_id", config=config)
    farm_id = get_setting("defaults.farm_id", config=config)

    if job_attachments_file_system is None:
        job_attachments_file_system = get_setting(
            "defaults.job_attachments_file_system", config=config
        )

    queue = deadline.get_queue(
        farmId=farm_id,
        queueId=queue_id,
    )
    print_function_callback(f"Submitting to Queue: {queue['displayName']}")

    create_job_args: Dict[str, Any] = {
        "farmId": farm_id,
        "queueId": queue_id,
        "template": file_contents,
        "templateType": file_type,
        "priority": 50,
    }

    storage_profile_id = get_setting("settings.storage_profile_id", config=config)
    storage_profile = None
    if storage_profile_id:
        create_job_args["storageProfileId"] = storage_profile_id
        storage_profile = api.get_storage_profile_for_queue(
            farm_id, queue_id, storage_profile_id, deadline
        )

    # The job parameters
    job_bundle_parameters = read_job_bundle_parameters(job_bundle_dir)

    asset_references_obj = read_yaml_or_json_object(
        job_bundle_dir, "asset_references", required=False
    )
    asset_references = AssetReferences.from_dict(asset_references_obj)

    if queue_parameter_definitions is None:
        queue_parameter_definitions = api.get_queue_parameter_definitions(
            farmId=farm_id, queueId=queue_id
        )

    parameters = merge_queue_job_parameters(
        queue_id=queue_id,
        job_parameters=job_bundle_parameters,
        queue_parameters=queue_parameter_definitions,
    )

    apply_job_parameters(
        job_parameters,
        job_bundle_dir,
        parameters,
        asset_references,
    )
    app_parameters_formatted, job_parameters_formatted = split_parameter_args(
        parameters, job_bundle_dir
    )

    # Hash and upload job attachments if there are any
    if asset_references and "jobAttachmentSettings" in queue:
        # Extend input_filenames with all the files in the input_directories
        missing_directories: set[str] = set()
        for directory in asset_references.input_directories:
            if not os.path.isdir(directory):
                if require_paths_exist:
                    missing_directories.add(directory)
                else:
                    logger.warning(
                        f"Input path '{directory}' does not exist. Adding to referenced paths."
                    )
                    asset_references.referenced_paths.add(directory)
                continue

            is_dir_empty = True
            for root, _, files in os.walk(directory):
                if not files:
                    continue
                is_dir_empty = False
                asset_references.input_filenames.update(
                    os.path.normpath(os.path.join(root, file)) for file in files
                )
            # Empty directories just become references since there's nothing to upload
            if is_dir_empty:
                logger.info(f"Input directory '{directory}' is empty. Adding to referenced paths.")
                asset_references.referenced_paths.add(directory)
        asset_references.input_directories.clear()

        if missing_directories:
            all_missing_directories = "\n\t".join(sorted(list(missing_directories)))
            misconfigured_directories_msg = (
                "Job submission contains misconfigured input directories and cannot be submitted."
                " All input directories must exist."
                f"\nNon-existent directories:\n\t{all_missing_directories}"
            )

            raise MisconfiguredInputsError(misconfigured_directories_msg)

        queue_role_session = api.get_queue_user_boto3_session(
            deadline=deadline,
            config=config,
            farm_id=farm_id,
            queue_id=queue_id,
            queue_display_name=queue["displayName"],
        )

        asset_manager = S3AssetManager(
            farm_id=farm_id,
            queue_id=queue_id,
            job_attachment_settings=JobAttachmentS3Settings(**queue["jobAttachmentSettings"]),
            session=queue_role_session,
        )

        upload_group = asset_manager.prepare_paths_for_upload(
            input_paths=sorted(asset_references.input_filenames),
            output_paths=sorted(asset_references.output_directories),
            referenced_paths=sorted(asset_references.referenced_paths),
            storage_profile=storage_profile,
            require_paths_exist=require_paths_exist,
        )
        if upload_group.asset_groups:
            if decide_cancel_submission_callback(upload_group):
                print_function_callback("Job submission canceled.")
                return None

            _, asset_manifests = _hash_attachments(
                asset_manager=asset_manager,
                asset_groups=upload_group.asset_groups,
                total_input_files=upload_group.total_input_files,
                total_input_bytes=upload_group.total_input_bytes,
                print_function_callback=print_function_callback,
                hashing_progress_callback=hashing_progress_callback,
            )

            attachment_settings = _upload_attachments(  # type: ignore
                asset_manager, asset_manifests, print_function_callback, upload_progress_callback
            )
            attachment_settings["fileSystem"] = JobAttachmentsFileSystem(
                job_attachments_file_system
            )
            create_job_args["attachments"] = attachment_settings

    create_job_args.update(app_parameters_formatted)

    if job_parameters_formatted:
        create_job_args["parameters"] = job_parameters_formatted

    if priority is not None:
        create_job_args["priority"] = priority
    if max_failed_tasks_count is not None:
        create_job_args["maxFailedTasksCount"] = max_failed_tasks_count
    if max_retries_per_task is not None:
        create_job_args["maxRetriesPerTask"] = max_retries_per_task

    if logging.DEBUG >= logger.getEffectiveLevel():
        logger.debug(json.dumps(create_job_args, indent=1))

    api.get_deadline_cloud_library_telemetry_client().record_event(
        event_type="com.amazon.rum.deadline.submission",
        event_details={"submitter_name": submitter_name},
    )

    create_job_response = deadline.create_job(**create_job_args)
    logger.debug(f"CreateJob Response {create_job_response}")

    if create_job_response and "jobId" in create_job_response:
        job_id = create_job_response["jobId"]
        print_function_callback("Waiting for Job to be created...")

        # If using the default config, set the default job id so it holds the
        # most-recently submitted job.
        if config is None:
            set_setting("defaults.job_id", job_id)

        def _default_create_job_result_callback() -> bool:
            return True

        if not create_job_result_callback:
            create_job_result_callback = _default_create_job_result_callback

        success, status_message = wait_for_create_job_to_complete(
            farm_id,
            queue_id,
            job_id,
            deadline,
            create_job_result_callback,
        )

        api.get_deadline_cloud_library_telemetry_client().record_event(
            event_type="com.amazon.rum.deadline.create_job", event_details={"is_success": success}
        )

        if not success:
            raise DeadlineOperationError(status_message)

        print_function_callback("Submitted job bundle:")
        print_function_callback(f"   {job_bundle_dir}")
        print_function_callback(status_message + f"\n{job_id}\n")

        return job_id
    else:
        raise DeadlineOperationError("CreateJob response was empty, or did not contain a Job ID.")


def wait_for_create_job_to_complete(
    farm_id: str,
    queue_id: str,
    job_id: str,
    deadline_client: BaseClient,
    continue_callback: Callable,
) -> Tuple[bool, str]:
    """
    Wait until a job exits the CREATE_IN_PROGRESS state.
    """

    delay_sec = 5  # Time to wait between checks in seconds.
    max_attempts = 60
    creating_statuses = {
        "CREATE_IN_PROGRESS",
    }
    failure_statuses = {"CREATE_FAILED"}

    for attempt in range(max_attempts):
        logger.debug(
            f"Waiting for creation of {job_id} to complete...attempt {attempt} of {max_attempts}"
        )

        if not continue_callback():
            raise CreateJobWaiterCanceled

        job = deadline_client.get_job(jobId=job_id, queueId=queue_id, farmId=farm_id)

        current_status = job["lifecycleStatus"] if "lifecycleStatus" in job else job["state"]
        if current_status in creating_statuses:
            time.sleep(delay_sec)
        elif current_status in failure_statuses:
            return False, job["lifecycleStatusMessage"]
        else:
            return True, job["lifecycleStatusMessage"]

    raise TimeoutError(
        f"Timed out after {delay_sec * max_attempts} seconds while waiting for Job to be created: {job_id}"
    )


@api.record_success_fail_telemetry_event(metric_name="cli_asset_upload")  # type: ignore
def _upload_attachments(
    asset_manager: S3AssetManager,
    manifests: List[AssetRootManifest],
    print_function_callback: Callable = lambda msg: None,
    upload_progress_callback: Optional[Callable] = None,
    config: Optional[ConfigParser] = None,
) -> Dict[str, Any]:
    """
    Starts the job attachments upload and handles the progress reporting callback.
    Returns the attachment settings from the upload.
    """

    def _default_update_upload_progress(upload_metadata: Dict[str, str]) -> bool:
        return True

    if not upload_progress_callback:
        upload_progress_callback = _default_update_upload_progress

    upload_summary, attachment_settings = asset_manager.upload_assets(
        manifests=manifests,
        on_uploading_assets=upload_progress_callback,
        s3_check_cache_dir=config_file.get_cache_directory(),
    )
    api.get_deadline_cloud_library_telemetry_client(config=config).record_upload_summary(
        upload_summary
    )

    print_function_callback("Upload Summary:")
    print_function_callback(textwrap.indent(str(upload_summary), "    "))

    return attachment_settings.to_dict()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Exceptions for the AWS Deadline Cloud Client Library.
"""


class DeadlineOperationError(Exception):
    """Error whose message gets printed verbatim by the cli handler"""


class CreateJobWaiterCanceled(Exception):
    """Error for when the waiter on CreateJob is interrupted"""


class UserInitiatedCancel(Exception):
    """Error for when the user requests cancelation"""


class NonValidInputError(Exception):
    """Error for when the user input is nonvalid"""


class ManifestOutdatedError(Exception):
    """Error for when local files are different from version captured in manifest"""



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
import concurrent.futures

import logging
import os
from pathlib import Path, PurePosixPath
from typing import Dict, List, Tuple
from deadline.client.cli._groups.click_logger import ClickLogger
from deadline.client.config import config_file
from deadline.client.exceptions import NonValidInputError
from deadline.job_attachments.asset_manifests.base_manifest import (
    BaseAssetManifest,
    BaseManifestPath,
)
from deadline.job_attachments.caches.hash_cache import HashCache
from deadline.job_attachments.models import AssetRootManifest, FileStatus, ManifestDiff
from deadline.job_attachments.upload import S3AssetManager


def diff_manifest(
    asset_manager: S3AssetManager,
    asset_root_manifest: AssetRootManifest,
    manifest: str,
    update: bool,
) -> List[(Tuple[FileStatus, BaseManifestPath])]:
    """
    Gets the file paths in specified manifest if the contents of file have changed since its last snapshot.
    Returns a list of FileStatus and BaseManifestPath
    """
    manifest_dir_name: str = os.path.basename(manifest)
    root_path: str = asset_root_manifest.root_path
    input_paths: List[Path] = []

    asset_manifest = asset_root_manifest.asset_manifest
    if asset_manifest is None:
        raise NonValidInputError("Manifest object not found, please check input manifest. ")

    for base_manifest_path in asset_manifest.paths:
        if base_manifest_path.path.startswith(manifest_dir_name):
            # skip the manifest folder, or else every upload will need an update after a previous change
            continue
        input_paths.append(Path(root_path, base_manifest_path.path))

    return find_file_with_status(
        asset_manager=asset_manager,
        input_paths=input_paths,
        root_path=root_path,
        update=update,
        statuses=[FileStatus.NEW, FileStatus.MODIFIED],
    )


def find_file_with_status(
    asset_manager: S3AssetManager,
    input_paths: List[Path],
    root_path: str,
    update: bool,
    statuses: List[FileStatus],
) -> List[(Tuple[FileStatus, BaseManifestPath])]:
    """
    Checks a manifest file, compares it to specified root directory or manifest of files with the local hash cache, and finds files that match the specified statuses.
    Returns a list of tuples containing the file information, and its corresponding file status.
    """
    cache_config: str = config_file.get_cache_directory()

    with HashCache(cache_config) as hash_cache:
        with concurrent.futures.ThreadPoolExecutor() as executor:
            futures = {
                executor.submit(
                    asset_manager._process_input_path,
                    path=path,
                    root_path=root_path,
                    hash_cache=hash_cache,
                    update=update,
                ): path
                for path in input_paths
            }
            status_paths: List[tuple] = []
            for future in concurrent.futures.as_completed(futures):
                (file_status, _, manifestPath) = future.result()
                if file_status in statuses:
                    status_paths.append((file_status, manifestPath))

            return status_paths


def compare_manifest(
    reference_manifest: BaseAssetManifest, compare_manifest: BaseAssetManifest
) -> List[(Tuple[FileStatus, BaseManifestPath])]:
    """
    Compares two manifests, reference_manifest acting as the base, and compare_manifest acting as manifest with changes.
    Returns a list of FileStatus and BaseManifestPath
    """
    reference_dict: Dict[str, BaseManifestPath] = {
        manifest_path.path: manifest_path for manifest_path in reference_manifest.paths
    }
    compare_dict: Dict[str, BaseManifestPath] = {
        manifest_path.path: manifest_path for manifest_path in compare_manifest.paths
    }

    differences: List[(Tuple[FileStatus, BaseManifestPath])] = []

    # Find new files
    for file_path, manifest_path in compare_dict.items():
        if file_path not in reference_dict:
            differences.append((FileStatus.NEW, manifest_path))
        elif reference_dict[file_path].hash != manifest_path.hash:
            differences.append((FileStatus.MODIFIED, manifest_path))
        else:
            differences.append((FileStatus.UNCHANGED, manifest_path))

    # Find deleted files
    for file_path, manifest_path in reference_dict.items():
        if file_path not in compare_dict:
            differences.append((FileStatus.DELETED, manifest_path))

    return differences


def _fast_file_list_to_manifest_diff(
    root: str,
    current_files: List[str],
    diff_manifest: BaseAssetManifest,
    logger: ClickLogger,
    return_root_relative_path: bool = True,
) -> List[Tuple[str, FileStatus]]:
    """
    Perform a fast difference of the current list of files to a previous manifest to diff against using time stamps and file sizes.
    :param root: Root folder of files to diff against.
    :param current_files: List of files to compare with.
    :param diff_manifest: Manifest containing files to diff against.
    :param return_root_relative_path: File Path to return, either relative to root or full.
    :param logger: logger.
    :return List[Tuple[str, FileStatus]]: List of Tuple containing the file path and FileStatus pair.
    """

    # Select either relative or absolut path for results.
    def select_path(full_path: str, relative_path: str, return_root_relative_path: bool):
        return relative_path if return_root_relative_path else full_path

    changed_paths: List[Tuple[str, FileStatus]] = []
    input_files_map: Dict[str, BaseManifestPath] = {}
    for input_file in diff_manifest.paths:
        # Normalize paths so we can compare different OSes
        normalized_path = Path(os.path.normpath(input_file.path)).as_posix()
        input_files_map[normalized_path] = input_file

    # Iterate for each file that we found in glob.
    root_relative_paths: List[str] = []
    for local_file in current_files:
        # Get the file's time stamp and size. We want to compare both.
        # From enabling CRT, sometimes timestamp update can fail.
        local_file_path = Path(local_file)
        file_stat = local_file_path.stat()

        # Compare the glob against the relative path we store in the manifest.
        # Save it to a list so we can look for deleted files.
        root_relative_path = str(PurePosixPath(*local_file_path.relative_to(root).parts).as_posix())
        root_relative_paths.append(root_relative_path)

        return_path = select_path(
            full_path=local_file,
            relative_path=root_relative_path,
            return_root_relative_path=return_root_relative_path,
        )
        if root_relative_path not in input_files_map:
            # This is a new file
            logger.echo(f"Found difference at: {root_relative_path}, Status: FileStatus.NEW")
            changed_paths.append((return_path, FileStatus.NEW))
        else:
            # This is a modified file, compare with manifest relative timestamp.
            input_file = input_files_map[root_relative_path]
            # Check file size first as it is easier to test. Usually modified files will also have size diff.
            if file_stat.st_size != input_file.size:
                changed_paths.append((return_path, FileStatus.MODIFIED))
                logger.echo(
                    f"Found size difference at: {root_relative_path}, Status: FileStatus.MODIFIED"
                )
            elif int(file_stat.st_mtime_ns // 1000) != input_file.mtime:
                changed_paths.append((return_path, FileStatus.MODIFIED))
                logger.echo(
                    f"Found time difference at: {root_relative_path}, Status: FileStatus.MODIFIED"
                )

    # Find deleted files. Manifest store files in relative form.
    for manifest_file_path in diff_manifest.paths:
        if manifest_file_path.path not in root_relative_paths:
            full_path = os.path.join(root, manifest_file_path.path)
            return_path = select_path(
                full_path=full_path,
                relative_path=manifest_file_path.path,
                return_root_relative_path=return_root_relative_path,
            )
            changed_paths.append((return_path, FileStatus.DELETED))
    return changed_paths


def pretty_print_cli(root: str, all_files: List[str], manifest_diff: ManifestDiff):
    """
    Prints to command line a formatted file tree structure with corresponding file statuses
    """

    # ASCII characters for the tree structure
    PIPE = "│"
    HORIZONTAL = "──"
    ELBOW = "└"
    TEE = "├"
    SPACE = "    "

    # ANSI escape sequences for colors
    COLORS = {
        "MODIFIED": "\033[93m",  # yellow
        "NEW": "\033[92m",  # green
        "DELETED": "\033[91m",  # red
        "UNCHANGED": "\033[90m",  # grey
        "RESET": "\033[0m",  # base color
        "DIRECTORY": "\033[80m",  # grey
    }

    # Tooltips:
    TOOLTIPS = {
        FileStatus.NEW: " +",  # added files
        FileStatus.DELETED: " -",  # deleted files
        FileStatus.MODIFIED: " M",  # modified files
        FileStatus.UNCHANGED: "",  # unchanged files
    }

    class ColorFormatter(logging.Formatter):
        def format(self, record):
            message = super().format(record)
            return f"{message}"

    # Configure logger
    formatter = ColorFormatter("")
    handler = logging.StreamHandler()
    handler.setFormatter(formatter)
    logger = logging.getLogger(__name__)
    logger.addHandler(handler)
    logger.setLevel(logging.INFO)
    logger.propagate = False

    def print_tree(directory_tree, prefix=""):
        sorted_entries = sorted(directory_tree.items())

        for i, (entry, subtree) in enumerate(sorted_entries, start=1):
            is_last_entry = i == len(sorted_entries)
            symbol = ELBOW + HORIZONTAL if is_last_entry else TEE + HORIZONTAL
            is_dir = isinstance(subtree, dict)
            color = COLORS["DIRECTORY"] if is_dir else COLORS[subtree.name]
            tooltip = TOOLTIPS[FileStatus.UNCHANGED] if is_dir else TOOLTIPS[subtree]

            message = f"{prefix}{symbol}{color}{entry}{tooltip}{COLORS['RESET']}{os.path.sep if is_dir else ''}"
            logger.info(message)

            if is_dir:
                new_prefix = prefix + (SPACE if is_last_entry else PIPE + SPACE)
                print_tree(subtree, new_prefix)

        if not directory_tree:
            symbol = ELBOW + HORIZONTAL
            message = f"{prefix}{symbol}{COLORS['UNCHANGED']}. {COLORS['RESET']}"
            logger.info(message)

    def get_file_status(file: str, manifest_diff: ManifestDiff):
        print(file)
        if file in manifest_diff.new:
            return FileStatus.NEW
        elif file in manifest_diff.modified:
            return FileStatus.MODIFIED
        elif file in manifest_diff.deleted:
            return FileStatus.DELETED
        else:
            # Default, not in any diff list.
            return FileStatus.UNCHANGED

    def build_directory_tree(all_files: List[str]) -> Dict[str, dict]:
        directory_tree: dict = {}

        def add_to_tree(path, status):
            parts = str(path).split(os.path.sep)
            current_level = directory_tree
            for i, part in enumerate(parts):
                if i == len(parts) - 1:
                    current_level[part] = status
                else:
                    current_level = current_level.setdefault(part, {})

        for file in all_files:
            print(f"{file} {root}")
            relative_path = str(Path(file).relative_to(root))
            add_to_tree(
                relative_path,
                get_file_status(relative_path, manifest_diff),
            )
        return directory_tree

    directory_tree = build_directory_tree(all_files)
    print_tree(directory_tree)
    logger.info("")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""UI Components for the Render Submitter"""
from __future__ import annotations

import logging
import os
import sys
from typing import Any, Dict, Optional

from qtpy.QtCore import QSize, Qt  # pylint: disable=import-error
from qtpy.QtGui import QKeyEvent  # pylint: disable=import-error
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QApplication,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QMessageBox,
    QPushButton,
    QScrollArea,
    QTabWidget,
    QVBoxLayout,
    QWidget,
)

from deadline.client.ui.dialogs.submit_job_progress_dialog import SubmitJobProgressDialog
from deadline.job_attachments.models import JobAttachmentS3Settings
from deadline.job_attachments.upload import S3AssetManager

from ... import api
from ..deadline_authentication_status import DeadlineAuthenticationStatus
from .. import block_signals
from ...config import get_setting
from ...config.config_file import str2bool
from ...exceptions import UserInitiatedCancel
from ...job_bundle import create_job_history_bundle_dir
from ...job_bundle.submission import AssetReferences
from ..widgets.deadline_authentication_status_widget import DeadlineAuthenticationStatusWidget
from ..widgets.job_attachments_tab import JobAttachmentsWidget
from ..widgets.shared_job_settings_tab import SharedJobSettingsWidget
from ..widgets.host_requirements_tab import HostRequirementsWidget
from . import DeadlineConfigDialog, DeadlineLoginDialog
from ._types import JobBundlePurpose

logger = logging.getLogger(__name__)

# initialize early so once the UI opens, things are already initialized
DeadlineAuthenticationStatus.getInstance()


class SubmitJobToDeadlineDialog(QDialog):
    """
    A widget containing all the standard tabs for submitting an AWS Deadline Cloud job.

    If you're using this dialog within an application and want it to stay in front,
    pass f=Qt.Tool, a flag that tells it to do that.

    Args:
        job_setup_widget_type (QWidget): The type of the widget for the job-specific settings.
        initial_job_settings (dataclass): A dataclass containing the initial job settings
        initial_shared_parameter_values (dict[str, Any]): A dict of parameter values {<name>, <value>, ...}
            to override default queue parameter values from the queue. For example,
            a Rez queue environment may have a default "" for the RezPackages parameter, but a Maya
            submitter would override that default with "maya-2023" or similar.
        auto_detected_attachments (FlatAssetReferences): The job attachments that were automatically detected
            from the input document/scene file or starting job bundle.
        attachments: (FlatAssetReferences): The job attachments that have been added to the job by the user.
        on_create_job_bundle_callback: A function to call when the dialog needs to create a Job Bundle. It
            is called with arguments (widget, job_bundle_dir, settings, queue_parameters, asset_references)
        parent: parent of the widget
        f: Qt Window Flags
        show_host_requirements_tab: Display the host requirements tab in dialog if set to True. Default
            to False.
        submitter_name: Override the default submitter_name value
    """

    def __init__(
        self,
        *,
        job_setup_widget_type: type[QWidget],
        initial_job_settings,
        initial_shared_parameter_values: dict[str, Any],
        auto_detected_attachments: AssetReferences,
        attachments: AssetReferences,
        on_create_job_bundle_callback,
        parent=None,
        f=Qt.WindowFlags(),
        show_host_requirements_tab=False,
        submitter_name: Optional[str] = None,
    ):
        # The Qt.Tool flag makes sure our widget stays in front of the main application window
        super().__init__(parent=parent, f=f)
        self.setWindowTitle("Submit to AWS Deadline Cloud")
        self.setMinimumSize(400, 400)

        self.job_settings_type = type(initial_job_settings)
        self.submitter_name = submitter_name or self.job_settings_type().submitter_name
        self.on_create_job_bundle_callback = on_create_job_bundle_callback
        self.create_job_response: Optional[Dict[str, Any]] = None
        self.job_history_bundle_dir: Optional[str] = None
        self.deadline_authentication_status = DeadlineAuthenticationStatus.getInstance()
        self.show_host_requirements_tab = show_host_requirements_tab

        self._build_ui(
            job_setup_widget_type,
            initial_job_settings,
            initial_shared_parameter_values,
            auto_detected_attachments,
            attachments,
        )

        self.gui_update_counter: Any = None
        self.refresh_deadline_settings()

    def sizeHint(self):
        return QSize(540, 600)

    def refresh(
        self,
        *,
        job_settings: Optional[Any] = None,
        auto_detected_attachments: Optional[AssetReferences] = None,
        attachments: Optional[AssetReferences] = None,
        load_new_bundle: bool = False,
    ):
        # Refresh the UI components
        self.refresh_deadline_settings()
        if (auto_detected_attachments is not None) or (attachments is not None):
            self.job_attachments.refresh_ui(auto_detected_attachments, attachments)

        if job_settings is not None:
            self.job_settings_type = type(job_settings)
            # Refresh shared job settings
            self.shared_job_settings.refresh_ui(job_settings, load_new_bundle)
            # Refresh job specific settings
            if hasattr(self.job_settings, "refresh_ui"):
                self.job_settings.refresh_ui(job_settings)

    def _build_ui(
        self,
        job_setup_widget_type,
        initial_job_settings,
        initial_shared_parameter_values,
        auto_detected_attachments: AssetReferences,
        attachments: AssetReferences,
    ):
        self.lyt = QVBoxLayout(self)
        self.lyt.setContentsMargins(5, 5, 5, 5)

        man_layout = QFormLayout()
        self.lyt.addLayout(man_layout)
        self.tabs = QTabWidget()
        self.lyt.addWidget(self.tabs)

        self._build_shared_job_settings_tab(initial_job_settings, initial_shared_parameter_values)
        self._build_job_settings_tab(job_setup_widget_type, initial_job_settings)
        self._build_job_attachments_tab(auto_detected_attachments, attachments)

        # Show host requirements only if requested by the constructor
        if self.show_host_requirements_tab:
            self._build_host_requirements_tab()

        self.auth_status_box = DeadlineAuthenticationStatusWidget(self)
        self.lyt.addWidget(self.auth_status_box)
        self.deadline_authentication_status.api_availability_changed.connect(
            self.refresh_deadline_settings
        )

        # Refresh the submit button enable state once queue parameter status changes
        self.shared_job_settings.valid_parameters.connect(self._set_submit_button_state)

        self.button_box = QDialogButtonBox(Qt.Horizontal)
        self.login_button = QPushButton("Login")
        self.login_button.clicked.connect(self.on_login)
        self.button_box.addButton(self.login_button, QDialogButtonBox.ResetRole)
        self.logout_button = QPushButton("Logout")
        self.logout_button.clicked.connect(self.on_logout)
        self.button_box.addButton(self.logout_button, QDialogButtonBox.ResetRole)
        self.settings_button = QPushButton("Settings...")
        self.settings_button.clicked.connect(self.on_settings_button_clicked)
        self.button_box.addButton(self.settings_button, QDialogButtonBox.ResetRole)
        self.submit_button = QPushButton("Submit")
        self.submit_button.clicked.connect(self.on_submit)
        self.button_box.addButton(self.submit_button, QDialogButtonBox.AcceptRole)
        self.export_bundle_button = QPushButton("Export bundle")
        self.export_bundle_button.clicked.connect(self.on_export_bundle)
        self.button_box.addButton(self.export_bundle_button, QDialogButtonBox.AcceptRole)

        self.lyt.addWidget(self.button_box)

    def _set_submit_button_state(self):
        # Enable/disable the Submit button based on whether the
        # AWS Deadline Cloud API is accessible and the farm+queue are configured.
        enable = (
            self.deadline_authentication_status.api_availability is True
            and get_setting("defaults.farm_id") != ""
            and get_setting("defaults.queue_id") != ""
            and self.shared_job_settings.is_queue_valid()
        )

        self.submit_button.setEnabled(enable)

        if not enable:
            self.submit_button.setToolTip(
                "Cannot submit job to Deadline Cloud. Nonvalid credentials or queue parameters."
            )
        else:
            self.submit_button.setToolTip("")

    def refresh_deadline_settings(self):
        # Enable/disable the Login and Logout buttons based on whether
        # the configured profile is for Deadline Cloud monitor
        self.login_button.setEnabled(
            self.deadline_authentication_status.creds_source
            == api.AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN
        )
        self.logout_button.setEnabled(
            self.deadline_authentication_status.creds_source
            == api.AwsCredentialsSource.DEADLINE_CLOUD_MONITOR_LOGIN
        )

        self._set_submit_button_state()

        self.shared_job_settings.deadline_cloud_settings_box.refresh_setting_controls(
            self.deadline_authentication_status.api_availability is True
        )
        # If necessary, this reloads the queue parameters
        self.shared_job_settings.refresh_queue_parameters()

    def keyPressEvent(self, event: QKeyEvent) -> None:
        """
        Override to capture any enter/return key presses so that the Submit
        button isn't "pressed" when the enter/return key is.
        """
        if event.key() == Qt.Key_Return or event.key() == Qt.Key_Enter:
            return
        super().keyPressEvent(event)

    def _build_shared_job_settings_tab(self, initial_job_settings, initial_shared_parameter_values):
        self.shared_job_settings_tab = QScrollArea()
        self.tabs.addTab(self.shared_job_settings_tab, "Shared job settings")
        self.shared_job_settings = SharedJobSettingsWidget(
            initial_settings=initial_job_settings,
            initial_shared_parameter_values=initial_shared_parameter_values,
            parent=self,
        )
        self.shared_job_settings.parameter_changed.connect(self.on_shared_job_parameter_changed)
        self.shared_job_settings_tab.setWidget(self.shared_job_settings)
        self.shared_job_settings_tab.setWidgetResizable(True)
        self.shared_job_settings.parameter_changed.connect(self.on_shared_job_parameter_changed)

    def _build_job_settings_tab(self, job_setup_widget_type, initial_job_settings):
        self.job_settings_tab = QScrollArea()
        self.tabs.addTab(self.job_settings_tab, "Job-specific settings")
        self.job_settings_tab.setWidgetResizable(True)

        self.job_settings = job_setup_widget_type(
            initial_settings=initial_job_settings, parent=self
        )
        self.job_settings_tab.setWidget(self.job_settings)
        if hasattr(self.job_settings, "parameter_changed"):
            self.job_settings.parameter_changed.connect(self.on_job_template_parameter_changed)

    def _build_job_attachments_tab(
        self, auto_detected_attachments: AssetReferences, attachments: AssetReferences
    ):
        self.job_attachments_tab = QScrollArea()
        self.tabs.addTab(self.job_attachments_tab, "Job attachments")
        self.job_attachments = JobAttachmentsWidget(
            auto_detected_attachments, attachments, parent=self
        )
        self.job_attachments_tab.setWidget(self.job_attachments)
        self.job_attachments_tab.setWidgetResizable(True)

    def _build_host_requirements_tab(self):
        self.host_requirements = HostRequirementsWidget()
        self.host_requirements_tab = QScrollArea()
        self.tabs.addTab(self.host_requirements_tab, "Host requirements")
        self.host_requirements_tab.setWidget(self.host_requirements)
        self.host_requirements_tab.setWidgetResizable(True)

    def on_shared_job_parameter_changed(self, parameter: dict[str, Any]):
        """
        Handles an edit to a shared job parameter, for example one of the
        queue parameters.

        When a queue parameter and a job template parameter have
        the same name, we update between them to keep them consistent.
        """
        try:
            if hasattr(self.job_settings, "set_parameter_value"):
                with block_signals(self.job_settings):
                    self.job_settings.set_parameter_value(parameter)
        except KeyError:
            # If there is no corresponding job template parameter,
            # just ignore it.
            pass

    def on_job_template_parameter_changed(self, parameter: dict[str, Any]):
        """
        Handles an edit to a job template parameter.

        When a queue parameter and a job template parameter have
        the same name, we update between them to keep them consistent.
        """
        try:
            with block_signals(self.shared_job_settings):
                self.shared_job_settings.set_parameter_value(parameter)
        except KeyError:
            # If there is no corresponding queue parameter,
            # just ignore it.
            pass

    def on_login(self):
        DeadlineLoginDialog.login(parent=self)
        self.refresh_deadline_settings()
        # This widget watches the auth files, but that does
        # not always catch a change so force a refresh here.
        self.deadline_authentication_status.refresh_status()

    def on_logout(self):
        api.logout()
        self.refresh_deadline_settings()
        # This widget watches the auth files, but that does
        # not always catch a change so force a refresh here.
        self.deadline_authentication_status.refresh_status()

    def on_settings_button_clicked(self):
        if DeadlineConfigDialog.configure_settings(parent=self):
            self.refresh_deadline_settings()

    def on_export_bundle(self):
        """
        Exports a Job Bundle, but does not submit the job.
        """
        # Retrieve all the settings into the dataclass
        settings = self.job_settings_type()
        self.shared_job_settings.update_settings(settings)
        self.job_settings.update_settings(settings)

        queue_parameters = self.shared_job_settings.get_parameters()

        asset_references = self.job_attachments.get_asset_references()

        # Save the bundle
        try:
            self.job_history_bundle_dir = create_job_history_bundle_dir(
                self.submitter_name, settings.name
            )

            if self.show_host_requirements_tab:
                requirements = self.host_requirements.get_requirements()
                self.on_create_job_bundle_callback(
                    self,
                    self.job_history_bundle_dir,
                    settings,
                    queue_parameters,
                    asset_references,
                    requirements,
                    purpose=JobBundlePurpose.EXPORT,
                )
            else:
                # Maintaining backward compatibility for submitters that do not support host_requirements yet
                self.on_create_job_bundle_callback(
                    self,
                    self.job_history_bundle_dir,
                    settings,
                    queue_parameters,
                    asset_references,
                    purpose=JobBundlePurpose.EXPORT,
                )

            logger.info(f"Saved the submission as a job bundle: {self.job_history_bundle_dir}")
            if sys.platform == "win32":
                # Open the directory in the OS's file explorer
                os.startfile(self.job_history_bundle_dir)
            QMessageBox.information(
                self,
                f"{self.submitter_name} job submission",
                f"Saved the submission as a job bundle:\n{self.job_history_bundle_dir}",
            )
            # Close the submitter window to signal the submission is done
            self.close()
        except Exception as exc:
            logger.exception("Error saving bundle")
            message = str(exc)
            QMessageBox.warning(self, f"{self.submitter_name} job submission", message)  # type: ignore[call-arg]

    def on_submit(self):
        """
        Perform a submission when the submit button is pressed
        """
        # Unset any cached response
        self.create_job_response = None

        # Retrieve all the settings into the dataclass
        settings = self.job_settings_type()
        self.shared_job_settings.update_settings(settings)
        self.job_settings.update_settings(settings)

        queue_parameters = self.shared_job_settings.get_parameters()

        asset_references = self.job_attachments.get_asset_references()

        job_progress_dialog = SubmitJobProgressDialog(parent=self)
        job_progress_dialog.show()
        QApplication.instance().processEvents()  # type: ignore[union-attr]

        # Submit the job
        try:
            deadline = api.get_boto3_client("deadline")

            self.job_history_bundle_dir = create_job_history_bundle_dir(
                self.submitter_name, settings.name
            )

            if self.show_host_requirements_tab:
                requirements = self.host_requirements.get_requirements()
                self.on_create_job_bundle_callback(
                    self,
                    self.job_history_bundle_dir,
                    settings,
                    queue_parameters,
                    asset_references,
                    requirements,
                    purpose=JobBundlePurpose.SUBMISSION,
                )
            else:
                # Maintaining backward compatibility for submitters that do not support host_requirements yet
                self.on_create_job_bundle_callback(
                    self,
                    self.job_history_bundle_dir,
                    settings,
                    queue_parameters,
                    asset_references,
                    purpose=JobBundlePurpose.SUBMISSION,
                )

            farm_id = get_setting("defaults.farm_id")
            queue_id = get_setting("defaults.queue_id")
            storage_profile_id = get_setting("settings.storage_profile_id")

            storage_profile = None
            if storage_profile_id:
                storage_profile = api.get_storage_profile_for_queue(
                    farm_id, queue_id, storage_profile_id, deadline
                )

            queue = deadline.get_queue(farmId=farm_id, queueId=queue_id)

            queue_role_session = api.get_queue_user_boto3_session(
                deadline=deadline,
                farm_id=farm_id,
                queue_id=queue_id,
                queue_display_name=queue["displayName"],
            )

            asset_manager: Optional[S3AssetManager] = None
            if "jobAttachmentSettings" in queue:
                asset_manager = S3AssetManager(
                    farm_id=farm_id,
                    queue_id=queue_id,
                    job_attachment_settings=JobAttachmentS3Settings(
                        **queue["jobAttachmentSettings"]
                    ),
                    session=queue_role_session,
                )

            api.get_deadline_cloud_library_telemetry_client().record_event(
                event_type="com.amazon.rum.deadline.submission",
                event_details={
                    "submitter_name": self.submitter_name,
                },
                from_gui=True,
            )

            self.create_job_response = job_progress_dialog.start_submission(
                farm_id,
                queue_id,
                storage_profile,
                self.job_history_bundle_dir,
                queue_parameters,
                asset_manager,
                deadline,
                auto_accept=str2bool(get_setting("settings.auto_accept")),
                require_paths_exist=self.job_attachments.get_require_paths_exist(),
            )
        except UserInitiatedCancel as uic:
            logger.info("Canceling submission.")
            QMessageBox.information(self, f"{self.submitter_name} job submission", str(uic))
            job_progress_dialog.close()
        except Exception as exc:
            logger.exception("error submitting job")
            api.get_deadline_cloud_library_telemetry_client().record_error(
                event_details={"exception_scope": "on_submit"},
                exception_type=str(type(exc)),
                from_gui=True,
            )
            QMessageBox.warning(self, f"{self.submitter_name} job submission", str(exc))  # type: ignore[call-arg]
            job_progress_dialog.close()

        if self.create_job_response:
            # Close the submitter window to signal the submission is done but
            # keep the standalone gui submitter open
            if self.submitter_name != "JobBundle":
                self.close()



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Module for defining a local cache file.
"""

import logging
import os
from abc import ABC
from threading import Lock
from typing import Optional

from ..exceptions import JobAttachmentsError

CONFIG_ROOT = ".deadline"
COMPONENT_NAME = "job_attachments"

logger = logging.getLogger("Deadline")


class CacheDB(ABC):
    """
    Abstract base class for connecting to a local SQLite cache database.

    This class is intended to always be used with a context manager to properly
    close the connection to the cache database.
    """

    def __init__(
        self, cache_name: str, table_name: str, create_query: str, cache_dir: Optional[str] = None
    ) -> None:
        if not cache_name or not table_name or not create_query:
            raise JobAttachmentsError("Constructor strings for CacheDB cannot be empty.")
        self.cache_name: str = cache_name
        self.table_name: str = table_name
        self.create_query: str = create_query

        try:
            # SQLite is included in Python installers, but might not exist if building python from source.
            import sqlite3  # noqa

            self.enabled = True
        except ImportError:
            logger.warn(f"SQLite was not found, {cache_name} will not be used.")
            self.enabled = False
            return

        if cache_dir is None:
            cache_dir = self.get_default_cache_db_file_dir()
        if cache_dir is None:
            raise JobAttachmentsError(
                f"No default cache path found. Please provide a directory for {self.cache_name}."
            )
        os.makedirs(cache_dir, exist_ok=True)
        self.cache_dir: str = os.path.join(cache_dir, f"{self.cache_name}.db")
        self.db_lock = Lock()

    def __enter__(self):
        """Called when entering the context manager."""
        if self.enabled:
            import sqlite3

            try:
                self.db_connection: sqlite3.Connection = sqlite3.connect(
                    self.cache_dir, check_same_thread=False
                )
            except sqlite3.OperationalError as oe:
                raise JobAttachmentsError(
                    f"Could not access cache file in {self.cache_dir}"
                ) from oe

            try:
                self.db_connection.execute(f"SELECT * FROM {self.table_name}")
            except Exception:
                # DB file doesn't have our table, so we need to create it
                logger.info(
                    f"No cache entries for the current library version were found. Creating a new cache for {self.cache_name}"
                )
                self.db_connection.execute(self.create_query)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Called when exiting the context manager."""
        if self.enabled:
            self.db_connection.close()

    @classmethod
    def get_default_cache_db_file_dir(cls) -> Optional[str]:
        """
        Gets the expected directory for the cache database file based on OS environment variables.
        If a directory cannot be found, defaults to the working directory.
        """
        default_path = os.environ.get("HOME")
        if default_path:
            default_path = os.path.join(default_path, CONFIG_ROOT, COMPONENT_NAME)
        return default_path



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
The AWS Deadline Cloud CLI interface.
"""

import logging
from logging import getLogger

import click

from .. import version
from ..config import get_setting, get_setting_default
from ._common import _PROMPT_WHEN_COMPLETE
from ._groups.bundle_group import cli_bundle
from ._groups.config_group import cli_config
from ._groups.auth_group import cli_auth
from ._groups.farm_group import cli_farm
from ._groups.fleet_group import cli_fleet
from ._groups.handle_web_url_command import cli_handle_web_url
from ._groups.job_group import cli_job
from ._groups.queue_group import cli_queue
from ._groups.worker_group import cli_worker
from ._groups.attachment_group import cli_attachment
from ._groups.manifest_group import cli_manifest

logger = getLogger(__name__)

CONTEXT_SETTINGS = dict(help_option_names=["-h", "--help"])
_DEADLINE_LOG_LEVELS = [
    "ERROR",
    "WARNING",
    "INFO",
    "DEBUG",
]  # Log Levels AWS Deadline Cloud Allows

# Set the default log level based on the setting, must do here so we can pass into the click option
_SETTING_LOG_LEVEL = get_setting("settings.log_level").upper()
_DEFAULT_LOG_LEVEL = get_setting_default("settings.log_level")
_CLI_DEFAULT_LOG_LEVEL = _DEFAULT_LOG_LEVEL
if _SETTING_LOG_LEVEL not in _DEADLINE_LOG_LEVELS:
    logger.warning(
        f"Log Level '{_SETTING_LOG_LEVEL}' not in {_DEADLINE_LOG_LEVELS}. Defaulting to {_DEFAULT_LOG_LEVEL}"
    )
else:
    _CLI_DEFAULT_LOG_LEVEL = _SETTING_LOG_LEVEL


@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(version=version, prog_name="deadline")
@click.option(
    "--log-level",
    type=click.Choice(_DEADLINE_LOG_LEVELS, case_sensitive=False),
    default=_CLI_DEFAULT_LOG_LEVEL,
    help="Set the logging level.",
)
@click.pass_context
def main(ctx: click.Context, log_level: str):
    """
    The AWS Deadline Cloud CLI provides functionality to interact with the AWS Deadline Cloud
    service.
    """
    logging.basicConfig(level=log_level)
    if log_level == "DEBUG":
        logger.debug("Debug logging is on")

    ctx.ensure_object(dict)
    # By default don't prompt when the operation is complete
    ctx.obj[_PROMPT_WHEN_COMPLETE] = False


main.add_command(cli_bundle)
main.add_command(cli_config)
main.add_command(cli_auth)
main.add_command(cli_farm)
main.add_command(cli_fleet)
main.add_command(cli_handle_web_url)
main.add_command(cli_job)
main.add_command(cli_queue)
main.add_command(cli_worker)

main.add_command(cli_attachment)
main.add_command(cli_manifest)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Contains dataclasses for holding UI parameter values, used by the widgets.
"""
from __future__ import annotations

import os
from dataclasses import dataclass, field

from ...job_bundle.parameters import JobParameter


@dataclass
class JobBundleSettings:  # pylint: disable=too-many-instance-attributes
    """
    Settings for the Job Bundle submitter dialog.
    """

    # Used in UI elements and when creating the job bundle directory
    submitter_name: str = field(default="JobBundle")

    # Shared settings
    name: str = field(default="Job bundle")
    description: str = field(default="")

    # Job Bundle settings
    input_job_bundle_dir: str = field(default="")
    parameters: list[JobParameter] = field(default_factory=list)

    # Whether to allow ability to "Load a different job bundle"
    browse_enabled: bool = field(default=False)


@dataclass
class CliJobSettings:  # pylint: disable=too-many-instance-attributes
    """
    Settings for a CLI Job.
    """

    # Used in UI elements and when creating the job bundle directory
    submitter_name: str = field(default="CLI")

    # Shared settings
    name: str = field(default="CLI job")
    description: str = field(default="")

    # CLI job settings
    bash_script_contents: str = field(
        default="""#!/usr/bin/env bash
echo "Data Dir is {{Param.DataDir}}"
cd "{{Param.DataDir}}"

echo "The file contents attached to this job:"
ls

echo "Running index {{Task.Param.Index}}"
sleep 35

# Generate an output file for this task
echo "Content for generated file {{Task.Param.Index}}" > task_output_file_{{Task.Param.Index}}.txt
"""
    )
    use_array_parameter: bool = field(default=True)
    array_parameter_name: str = field(default="Index")
    array_parameter_values: str = field(default="1-5")
    data_dir: str = field(default=os.path.join("~", "CLIJobData"))
    file_format: str = field(default="YAML")



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

import json
from pathlib import Path
from typing import List, Optional
from deadline.client.exceptions import NonValidInputError
from deadline.job_attachments.models import GlobConfig


def _process_glob_inputs(glob_arg_input: str) -> GlobConfig:
    """
    Helper function to process glob inputs.
    glob_input: String, can represent a json, filepath, or general include glob syntax.
    """

    # Default Glob config.
    glob_config = GlobConfig()
    if glob_arg_input is None or len(glob_arg_input) == 0:
        # Not configured, or not passed in.
        return glob_config

    try:
        input_as_path = Path(glob_arg_input)
        if input_as_path.is_file():
            # Read the file so it can be parsed as JSON.
            with open(glob_arg_input) as f:
                glob_arg_input = f.read()
    except Exception:
        # If this cannot be processed as a file, try it as JSON.
        pass

    try:
        # Parse the input as JSON, default to Glob Config defaults.
        input_as_json = json.loads(glob_arg_input)
        glob_config.include_glob = input_as_json.get(GlobConfig.INCLUDE, glob_config.include_glob)
        glob_config.exclude_glob = input_as_json.get(GlobConfig.EXCLUDE, glob_config.exclude_glob)
    except Exception:
        # This is not a JSON blob, bad input.
        raise NonValidInputError(f"Glob input {glob_arg_input} cannot be deserialized as JSON")

    return glob_config


def _glob_paths(
    path: str, include: List[str] = ["**/*"], exclude: Optional[List[str]] = None
) -> List[str]:
    """
    Glob routine that supports Unix style pathname pattern expansion for includes and excludes.
    This function will recursively list all files of path, including all files globbed by include and removing all files marked by exclude.
    path: Root path to glob.
    include: Optional, pattern syntax for files to include.
    exclude: Optional, pattern syntax for files to exclude.
    return: List of files found based on supplied glob patterns.
    """
    include_files: List[Optional[str]] = []
    for input_glob in include:
        include_files.extend(
            [str(path) if path.is_file() else None for path in Path(path).glob(input_glob)]
        )
    include_files = list(filter(None, include_files))

    if exclude:
        exclude_files = []
        for exclude_glob in exclude:
            exclude_files.extend(
                [str(path) if path.is_file() else None for path in Path(path).glob(exclude_glob)]
            )

        # use sets to remove duplicates and remove all excluded file.
        exclude_files = list(filter(None, exclude_files))
        include_files = list(set(include_files) - set(exclude_files))
    return include_files  # type: ignore[return-value]



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
"""
Provides a modal dialog box for the submission progress when submitting to
AWS Deadline Cloud
"""
from __future__ import annotations

import json
import logging
import os
import threading
import textwrap
from typing import Any, Dict, List, Optional, cast

from botocore.client import BaseClient  # type: ignore[import]
from qtpy.QtCore import Qt, Signal
from qtpy.QtGui import QCloseEvent
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QApplication,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGroupBox,
    QLabel,
    QMessageBox,
    QPushButton,
    QProgressBar,
    QTextEdit,
    QVBoxLayout,
    QWidget,
)

from deadline.client import api
from deadline.client.exceptions import (
    CreateJobWaiterCanceled,
    DeadlineOperationError,
    UserInitiatedCancel,
)
from deadline.client.config import set_setting, config_file
from deadline.client.job_bundle.loader import (
    read_yaml_or_json,
    read_yaml_or_json_object,
    validate_directory_symlink_containment,
)
from deadline.client.job_bundle.parameters import (
    JobParameter,
    apply_job_parameters,
    merge_queue_job_parameters,
    read_job_bundle_parameters,
)
from deadline.client.job_bundle.submission import (
    AssetReferences,
    split_parameter_args,
)
from deadline.job_attachments.exceptions import AssetSyncCancelledError, MisconfiguredInputsError
from deadline.job_attachments.models import (
    AssetRootGroup,
    AssetRootManifest,
    AssetUploadGroup,
    StorageProfile,
)
from deadline.job_attachments.progress_tracker import ProgressReportMetadata, SummaryStatistics
from deadline.job_attachments.upload import S3AssetManager
from deadline.job_attachments._utils import _human_readable_file_size

__all__ = ["SubmitJobProgressDialog"]

logger = logging.getLogger(__name__)


class SubmitJobProgressDialog(QDialog):
    """
    A modal dialog box for the submission progress while submitting a job bundle
    to AWS Deadline Cloud.
    """

    # These signals are sent when the background threads raise an exception.
    hashing_thread_exception = Signal(BaseException)
    upload_thread_exception = Signal(BaseException)
    create_job_thread_exception = Signal(BaseException)

    # These signals are sent when the background threads succeed.
    hashing_thread_succeeded = Signal(SummaryStatistics, list)
    upload_thread_succeeded = Signal(SummaryStatistics, dict)
    create_job_thread_succeeded = Signal(bool, str)

    # These signals are sent when the progress reporting callbacks are called
    # from job attachments during hashing/uploading.
    hashing_thread_progress_report = Signal(ProgressReportMetadata)
    upload_thread_progress_report = Signal(ProgressReportMetadata)

    def __init__(self, parent: QWidget) -> None:
        super().__init__(parent=parent)
        self._continue_submission = True
        self._submission_complete = False
        self._create_job_args: Dict[str, Any] = {}
        self._create_job_response: Dict[str, Any] = {}
        self.__hashing_thread: Optional[threading.Thread] = None
        self.__upload_thread: Optional[threading.Thread] = None
        self.__create_job_thread: Optional[threading.Thread] = None

        self._build_ui()

    def start_submission(
        self,
        farm_id: str,
        queue_id: str,
        storage_profile: Optional[StorageProfile],
        job_bundle_dir: str,
        queue_parameters: list[JobParameter],
        asset_manager: Optional[S3AssetManager],
        deadline_client: BaseClient,
        auto_accept: bool = False,
        require_paths_exist: bool = False,
    ) -> Optional[Dict[str, Any]]:
        """
        Starts a submission. Returns the response from calling create job. If an error occurs
        or the submission is canceled then None is returned.

        Args:
            farm_id (str): Id of the farm to submit to
            queue_id (str): Id of the queue to submit to
            storage_profile (StorageProfile): the storage profile to associate
                with the job.
            job_bundle_dir (str): Path to the folder containing the job bundle to
                submit.
            asset_manager (S3AssetManager): A job attachments S3AssetManager
                configured for the farm/queue to submit to
            deadline_client (BaseClient): A boto client for AWS Deadline Cloud
            auto_accept (bool, default False): Flag for whether any confirmation
                prompts should automatically be accepted.
        """
        self._farm_id = farm_id
        self._queue_id = queue_id
        self._storage_profile = storage_profile
        self._job_bundle_dir = job_bundle_dir
        self._queue_parameters = queue_parameters
        self._asset_manager = asset_manager
        self._deadline_client = deadline_client
        self._auto_accept = auto_accept
        self._require_paths_exist = require_paths_exist

        self._start_submission()
        return self.exec_()

    def _build_ui(self):
        """Builds up the Dialog UI"""
        # Remove help button from title bar
        self.setWindowFlags(self.windowFlags() & ~Qt.WindowContextHelpButtonHint)
        self.lyt = QVBoxLayout(self)
        self.lyt.setContentsMargins(5, 10, 5, 5)
        self.setMinimumWidth(450)

        self.status_label = QLabel("Preparing files...")
        self.status_label.setMargin(5)
        self.hashing_progress = JobAttachmentsProgressWidget(
            initial_message="Preparing for hashing...", title="Hashing progress", parent=self
        )
        self.upload_progress = JobAttachmentsProgressWidget(
            initial_message="Preparing for upload...", title="Upload progress", parent=self
        )
        self.summary_edit = QTextEdit()
        self.summary_edit.setVisible(False)
        self.summary_edit.setReadOnly(True)
        self.button_box = QDialogButtonBox(Qt.Horizontal)
        self.button_box.setStandardButtons(QDialogButtonBox.Cancel)

        self.lyt.setAlignment(Qt.AlignTop)
        self.lyt.addWidget(self.status_label)
        self.lyt.addWidget(self.hashing_progress)
        self.lyt.addWidget(self.upload_progress)
        self.lyt.addWidget(self.summary_edit)
        self.lyt.addWidget(self.button_box)

        self.setWindowTitle("AWS Deadline Cloud submission")

        self.hashing_thread_progress_report.connect(self.handle_hashing_thread_progress_report)
        self.hashing_thread_succeeded.connect(self.handle_hashing_thread_succeeded)
        self.hashing_thread_exception.connect(self.handle_thread_exception)

        self.upload_thread_progress_report.connect(self.handle_upload_thread_progress_report)
        self.upload_thread_succeeded.connect(self.handle_upload_thread_succeeded)
        self.upload_thread_exception.connect(self.handle_thread_exception)

        self.create_job_thread_succeeded.connect(self.handle_create_job_thread_succeeded)
        self.create_job_thread_exception.connect(self.handle_thread_exception)

        self.button_box.accepted.connect(self.accept)
        self.button_box.rejected.connect(self.close)

    def _start_submission(self):
        """
        Starts building up the arguments to pass to create job. If there are any
        job attachments the hashing process will be started. If there are no job
        attachments then create job will be called without calling job attachments.
        """

        # Ensure the job bundle doesn't contain files that resolve outside of the bundle directory
        validate_directory_symlink_containment(self._job_bundle_dir)

        # Read in the job template
        file_contents, file_type = read_yaml_or_json(
            self._job_bundle_dir, "template", required=True
        )

        self._create_job_args["farmId"] = self._farm_id
        self._create_job_args["queueId"] = self._queue_id
        self._create_job_args["template"] = file_contents
        self._create_job_args["templateType"] = file_type

        if self._storage_profile:
            self._create_job_args["storageProfileId"] = self._storage_profile.storageProfileId

        # The job parameters
        job_bundle_parameters = read_job_bundle_parameters(self._job_bundle_dir)

        asset_references_obj = read_yaml_or_json_object(
            self._job_bundle_dir, "asset_references", required=False
        )
        self.asset_references = AssetReferences.from_dict(asset_references_obj)

        parameter_definitions = merge_queue_job_parameters(
            queue_id=self._queue_id,
            job_parameters=job_bundle_parameters,
            queue_parameters=self._queue_parameters,
        )

        apply_job_parameters(
            [],
            self._job_bundle_dir,
            parameter_definitions,
            self.asset_references,
        )

        app_parameters_formatted, job_parameters_formatted = split_parameter_args(
            parameter_definitions, self._job_bundle_dir
        )

        self._create_job_args.update(app_parameters_formatted)

        if job_parameters_formatted:
            self._create_job_args["parameters"] = job_parameters_formatted

        if self._asset_manager and (
            self.asset_references.input_filenames
            or self.asset_references.input_directories
            or self.asset_references.output_directories
        ):
            # Extend input_filenames with all the files in the input_directories
            missing_directories: set[str] = set()
            for directory in self.asset_references.input_directories:
                if not os.path.isdir(directory):
                    if self._require_paths_exist:
                        missing_directories.add(directory)
                    else:
                        logging.warning(
                            f"Input directory '{directory}' does not exist. Adding to referenced paths."
                        )
                        self.asset_references.referenced_paths.add(directory)
                    continue

                is_dir_empty = True
                for root, _, files in os.walk(directory):
                    if not files:
                        continue
                    is_dir_empty = False
                    self.asset_references.input_filenames.update(
                        os.path.normpath(os.path.join(root, file)) for file in files
                    )
                # Empty directories just become references since there's nothing to upload
                if is_dir_empty:
                    logging.info(
                        f"Input directory '{directory}' is empty. Adding to referenced paths."
                    )
                    self.asset_references.referenced_paths.add(directory)
            self.asset_references.input_directories.clear()

            if missing_directories:
                sample_size = 3
                misconfigured_directories_msg = (
                    "Job submission contains misconfigured input directories and cannot be submitted."
                    " All input directories must exist."
                )

                missing_directory_list = sorted(list(missing_directories))
                sample_of_missing_directories = "\n\t".join(missing_directory_list[:sample_size])
                sample_of_misconfigured_inputs = (
                    f"\nNon-existent directories:\n\t{sample_of_missing_directories}\n"
                )
                all_missing_directories = "\n\t".join(missing_directory_list)
                all_misconfigured_inputs = (
                    f"\nNon-existent directories:\n\t{all_missing_directories}"
                )

                logging.error(misconfigured_directories_msg + all_misconfigured_inputs)
                if len(missing_directories) > sample_size:
                    misconfigured_directories_msg += (
                        " Check logs for all occurrences, here's a sample:\n"
                    )
                misconfigured_directories_msg += f"{sample_of_misconfigured_inputs}"

                raise MisconfiguredInputsError(misconfigured_directories_msg)

            upload_group = self._asset_manager.prepare_paths_for_upload(
                input_paths=sorted(self.asset_references.input_filenames),
                output_paths=sorted(self.asset_references.output_directories),
                referenced_paths=sorted(self.asset_references.referenced_paths),
                storage_profile=self._storage_profile,
                require_paths_exist=self._require_paths_exist,
            )
            # If we find any Job Attachments, start a background thread
            if upload_group.asset_groups:
                if not self._confirm_asset_references_outside_storage_profile(upload_group):
                    raise UserInitiatedCancel("Submission canceled.")

                self._start_hashing(
                    upload_group.asset_groups,
                    upload_group.total_input_files,
                    upload_group.total_input_bytes,
                )
                return

        self.hashing_progress.setVisible(False)
        self.upload_progress.setVisible(False)
        self._start_create_job()

    def _hashing_background_thread(
        self,
        asset_groups: list[AssetRootGroup],
        total_input_files: int,
        total_input_bytes: int,
    ) -> None:
        """
        This function gets started in a background thread to start the hashing
        of any job attachments.
        """
        try:

            def _update_hash_progress(hashing_metadata: ProgressReportMetadata) -> bool:
                self.hashing_thread_progress_report.emit(hashing_metadata)
                return self._continue_submission

            logger.info("Hashing job attachments files...")

            # This thread is only started if self._asset_manager is set.
            hashing_summary, manifests = cast(
                S3AssetManager, self._asset_manager
            ).hash_assets_and_create_manifest(
                asset_groups=asset_groups,
                total_input_files=total_input_files,
                total_input_bytes=total_input_bytes,
                hash_cache_dir=config_file.get_cache_directory(),
                on_preparing_to_submit=_update_hash_progress,
            )

            logger.info("Finished hashing job attachments files.")

            self.hashing_thread_succeeded.emit(hashing_summary, manifests)
        except AssetSyncCancelledError as e:
            # If it wasn't canceled, send the exception to the dialog
            if self._continue_submission:
                self.hashing_thread_exception.emit(e)
            else:
                logger.info("Job attachments hashing canceled.")
        except Exception as e:
            # Send the exception to the dialog
            self.hashing_thread_exception.emit(e)

    @api.record_success_fail_telemetry_event(metric_name="gui_asset_upload")  # type: ignore
    def _upload_background_thread(self, manifests: List[AssetRootManifest]) -> None:
        """
        This function gets started in a background thread to start the upload
        of any job attachments.
        """
        try:

            def _update_upload_progress(upload_metadata: ProgressReportMetadata) -> bool:
                self.upload_thread_progress_report.emit(upload_metadata)
                return self._continue_submission

            logger.info("Uploading job attachments files...")

            # This thread is only started if self._asset_manager is set.
            upload_summary, attachment_settings = cast(
                S3AssetManager, self._asset_manager
            ).upload_assets(
                manifests=manifests,
                on_uploading_assets=_update_upload_progress,
                s3_check_cache_dir=config_file.get_cache_directory(),
                manifest_write_dir=self._job_bundle_dir,
            )

            logger.info("Finished uploading job attachments files.")

            self.upload_thread_succeeded.emit(upload_summary, attachment_settings.to_dict())
        except AssetSyncCancelledError as e:
            # If it wasn't canceled, send the exception to the dialog
            if self._continue_submission:
                self.hashing_thread_exception.emit(e)
            else:
                logger.info("Job attachments upload canceled.")
        except Exception as e:
            # Send the exception to the dialog
            self.hashing_thread_exception.emit(e)

    def _create_job_background_thread(self) -> None:
        """
        This function gets started in a background thread to call CreateJob.
        """
        try:

            def _continue_create_job_wait() -> bool:
                return self._continue_submission

            logger.info("Waiting for job to be created...")

            success = False

            logger.debug(json.dumps(self._create_job_args, indent=1))
            self._create_job_response = self._deadline_client.create_job(**self._create_job_args)
            logger.debug(f"CreateJob Response {self._create_job_response}")

            if self._create_job_response and "jobId" in self._create_job_response:
                job_id = self._create_job_response["jobId"]

                # Set the default job id so it holds the most-recently submitted job.
                set_setting("defaults.job_id", job_id)

                success, message = api.wait_for_create_job_to_complete(
                    self._farm_id,
                    self._queue_id,
                    job_id,
                    self._deadline_client,
                    _continue_create_job_wait,
                )
                message += f"\n{job_id}\n"
            else:
                message = "CreateJob response was empty, or did not contain a job ID."
            if success:
                self.create_job_thread_succeeded.emit(success, message)
            else:
                self.create_job_thread_exception.emit(DeadlineOperationError(message))
        except CreateJobWaiterCanceled as e:
            # If it wasn't canceled, send the exception to the dialog
            if self._continue_submission:
                self.create_job_thread_exception.emit(e)
            else:
                logger.info("Wait for CreateJob result canceled.")
        except Exception as e:
            # Send the exception to the dialog
            self.create_job_thread_exception.emit(e)

    def _start_hashing(
        self,
        asset_groups: list[AssetRootGroup],
        total_input_files: int,
        total_input_bytes: int,
    ) -> None:
        """
        Starts the background hashing thread.
        """
        self.status_label.setText("Hashing job attachments...")
        self.__hashing_thread = threading.Thread(
            target=self._hashing_background_thread,
            name="AWS Deadline Cloud hashing background thread",
            args=(asset_groups, total_input_files, total_input_bytes),
        )
        self.__hashing_thread.start()

    def _start_upload(self, asset_manifests: List[AssetRootManifest]) -> None:
        """
        Starts the background upload thread.
        """
        self.status_label.setText("Uploading job attachments...")
        self.__upload_thread = threading.Thread(
            target=self._upload_background_thread,
            name="AWS Deadline Cloud upload background thread",
            args=(asset_manifests,),
        )
        self.__upload_thread.start()

    def _start_create_job(self) -> None:
        """
        Starts the background thread to call CreateJob.
        """
        self.status_label.setText("Waiting for job to be created...")
        self.__create_job_thread = threading.Thread(
            target=self._create_job_background_thread,
            name="AWS Deadline Cloud CreateJob background thread",
        )
        self.__create_job_thread.start()

    def handle_hashing_thread_progress_report(
        self, progress_metadata: ProgressReportMetadata
    ) -> None:
        """
        Handles the signal sent from the background threads when reporting
        hashing progress. Sets the progress bar in the dialog based on
        the callback progress data from job attachments.
        """
        self.hashing_progress.progress_bar.setValue(int(progress_metadata.progress))
        self.hashing_progress.progress_message.setText(progress_metadata.progressMessage)

    def handle_upload_thread_progress_report(
        self, progress_metadata: ProgressReportMetadata
    ) -> None:
        """
        Handles the signal sent from the background threads when reporting
        upload progress. Sets the progress bar in the dialog based on
        the callback progress data from job attachments.
        """
        self.upload_progress.progress_bar.setValue(int(progress_metadata.progress))
        self.upload_progress.progress_message.setText(progress_metadata.progressMessage)

    def handle_hashing_thread_succeeded(
        self,
        hashing_summary: SummaryStatistics,
        asset_manifests: List[AssetRootManifest],
    ) -> None:
        """
        Handles the signal sent from the background hashing thread when the
        hashing process has completed.
        """
        api.get_deadline_cloud_library_telemetry_client().record_hashing_summary(
            hashing_summary, from_gui=True
        )
        self.summary_edit.setText(
            f"\nHashing summary:\n{textwrap.indent(str(hashing_summary), '    ')}"
        )
        self._start_upload(asset_manifests)

    def handle_upload_thread_succeeded(
        self, upload_summary: SummaryStatistics, attachment_settings: Any
    ) -> None:
        """
        Handles the signal sent from the background upload thread when the upload
        has finished.
        """
        if attachment_settings:
            self._create_job_args["attachments"] = attachment_settings
            self._create_job_args["attachments"]["fileSystem"] = config_file.get_setting(
                "defaults.job_attachments_file_system"
            )

        api.get_deadline_cloud_library_telemetry_client().record_upload_summary(
            upload_summary, from_gui=True
        )
        self.summary_edit.setText(
            f"{self.summary_edit.toPlainText()}"
            + f"\nUpload summary:\n{textwrap.indent(str(upload_summary), '    ')}"
        )

        self._start_create_job()

    def handle_create_job_thread_succeeded(self, success: bool, status_message: str) -> None:
        """
        Handles the signal sent from the background CreateJob thread when the
        job creation has finished.
        """
        api.get_deadline_cloud_library_telemetry_client().record_event(
            event_type="com.amazon.rum.deadline.create_job",
            event_details={"is_success": success},
            from_gui=True,
        )

        if success:
            self._submission_complete = True
            self.status_label.setText("Submission complete")
            self.button_box.setStandardButtons(QDialogButtonBox.Ok)
            self.button_box.button(QDialogButtonBox.Ok).setDefault(True)
        else:
            self.status_label.setText("Submission error")
            self.button_box.setStandardButtons(QDialogButtonBox.Close)
            self.button_box.button(QDialogButtonBox.Close).setDefault(True)

        self.summary_edit.setText(f"{status_message} {self.summary_edit.toPlainText()}")
        self.summary_edit.setVisible(True)
        self.adjustSize()

    def handle_thread_exception(self, e: BaseException) -> None:
        """
        Handles the signal sent from the background threads when an exception is
        thrown.
        """
        self.hashing_progress.setVisible(False)
        self.upload_progress.setVisible(False)
        self.status_label.setVisible(False)
        self.button_box.setStandardButtons(QDialogButtonBox.Close)
        self.summary_edit.setText(f"Error occurred: {str(e)}")
        self.summary_edit.setVisible(True)
        self.adjustSize()
        logger.exception(e, exc_info=(type(e), e, e.__traceback__))

    def _confirm_asset_references_outside_storage_profile(
        self, upload_group: AssetUploadGroup
    ) -> bool:
        """
        Creates a dialog to prompt the user to confirm that they want to proceed
        with uploading when files were found outside of the configured storage profile locations.
        """
        message_text = (
            f"Job submission contains {upload_group.total_input_files} input files totaling {_human_readable_file_size(upload_group.total_input_bytes)}. "
            " All input files will be uploaded to S3 if they are not already present in the job attachments bucket."
        )
        warning_message = ""
        for group in upload_group.asset_groups:
            if not group.file_system_location_name:
                warning_message += f"\n\nUnder the directory '{group.root_path}':"
                warning_message += (
                    f"\n\t{len(group.inputs)} input file{'' if len(group.inputs) == 1 else 's'}"
                    if len(group.inputs) > 0
                    else ""
                )
                warning_message += (
                    f"\n\t{len(group.outputs)} output director{'y' if len(group.outputs) == 1 else 'ies'}"
                    if len(group.outputs) > 0
                    else ""
                )
                warning_message += (
                    f"\n\t{len(group.references)} referenced file{'' if len(group.references) == 1 else 's'} and/or director{'y' if len(group.outputs) == 1 else 'ies'}"
                    if len(group.references) > 0
                    else ""
                )

        # Exit early if we've set auto accept and there are no warnings
        if not warning_message and self._auto_accept:
            return True

        # Build the UI
        message_box = QMessageBox(self)
        if warning_message:
            if self._storage_profile:
                fs_locations_text = "\n\t".join(
                    [fs_location.path for fs_location in self._storage_profile.fileSystemLocations]
                )
                message_text += f"\n\nFiles were specified outside of the configured storage profile location(s):\n{fs_locations_text}\n"
            else:
                message_text += "\n\nNo storage profile locations are configured for this queue."
            message_text += (
                "\nPlease confirm that you intend to submit a job that uses files from the following directories:"
                f"{warning_message}\n\n"
                "To permanently remove this warning you must only use files located within a storage profile location."
            )
            message_box.setIcon(QMessageBox.Warning)

        message_box.setText(message_text)
        message_box.setStandardButtons(QMessageBox.Ok | QMessageBox.Cancel)
        message_box.setDefaultButton(QMessageBox.Ok)

        if not warning_message:
            # If we don't have any warnings, add the "Do not ask again" button that acts like 'OK' but sets the config
            # setting to always auto-accept similar prompts in the future.
            dont_ask_button = QPushButton("Do not ask again", self)
            dont_ask_button.clicked.connect(lambda: set_setting("settings.auto_accept", "true"))
            message_box.addButton(dont_ask_button, QMessageBox.ActionRole)

        message_box.setWindowTitle("Job attachments valid files confirmation")
        selection = message_box.exec()

        return selection != QMessageBox.Cancel

    def closeEvent(self, event: QCloseEvent) -> None:
        """
        Overrides the closeEvent function to shutdown any running threads before
        closing the dialog. If the submission is complete then any button, even
        'X', should result in the dialog being accepted.
        """
        if self._submission_complete:
            self.accept()
        else:
            logger.info("Canceling submission...")
            self.status_label.setText("Canceling submission...")
            self.hashing_progress.setVisible(False)
            self.upload_progress.setVisible(False)
            self.adjustSize()
            self._continue_submission = False
            self._shutdown_threads()
            super().closeEvent(event)

    def _shutdown_threads(self) -> None:
        """Closes any threads. Used before canceling/closing"""

        threads = (self.__hashing_thread, self.__upload_thread, self.__create_job_thread)

        for thread in threads:
            if thread:
                while thread.is_alive():
                    QApplication.instance().processEvents()  # type: ignore[union-attr]

    def exec_(self) -> Optional[Dict[str, Any]]:  # type: ignore[override]
        """
        Runs the modal job progress dialog, returns the response from calling
        create job if the dialog was accepted. Otherwise returns None
        """
        if super().exec_() == QDialog.Accepted:
            return self._create_job_response
        return None


class JobAttachmentsProgressWidget(QGroupBox):
    """
    UI element to group job attachments progress bar with a status message.
    """

    def __init__(self, *, initial_message: str, title: str, parent: Optional[QWidget] = None):
        super().__init__(parent=parent, title=title)
        self.initial_message = initial_message

        self._build_ui()

    def _build_ui(self):
        self.layout = QFormLayout(self)
        self.layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)

        self.progress_bar = QProgressBar()
        self.progress_message = QLabel(self.initial_message)

        self.layout.addWidget(self.progress_bar)
        self.layout.addWidget(self.progress_message)



# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.

"""
Provides a widget to place in AWS Deadline Cloud submitter dialogs, that shows
the current status of AWS Deadline Cloud authentication and API.
The current status is handled by DeadlineAuthenticationStatus.
"""
from logging import getLogger
from typing import Optional

from qtpy.QtCore import Qt
from qtpy.QtWidgets import (  # pylint: disable=import-error; type: ignore
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QWidget,
)

from ... import api
from ..deadline_authentication_status import DeadlineAuthenticationStatus

logger = getLogger(__name__)


class DeadlineAuthenticationStatusWidget(QWidget):
    """
    A Widget that displays status information about AWS Deadline Cloud
    authentication from a DeadlineAuthenticationStatus object.
    """

    def __init__(self, parent=None) -> None:
        super().__init__(parent=parent)

        layout = QHBoxLayout(self)

        self.creds_source_group = AuthenticationStatusGroup(title="Credential source", parent=self)
        layout.addWidget(self.creds_source_group)
        self.auth_status_group = AuthenticationStatusGroup(
            title="Authentication status", parent=self
        )
        layout.addWidget(self.auth_status_group)
        self.deadline_authorized_group = AuthenticationStatusGroup(
            title="AWS Deadline Cloud API", parent=self
        )
        layout.addWidget(self.deadline_authorized_group)

        self._status = DeadlineAuthenticationStatus.getInstance()
        self._status.creds_source_changed.connect(self._creds_source_changed)
        self._status.auth_status_changed.connect(self._auth_status_changed)
        self._status.api_availability_changed.connect(self._api_availability_changed)

        # Update with current values
        self._creds_source_changed()
        self._auth_status_changed()
        self._api_availability_changed()

    def _creds_source_changed(self) -> None:
        if self._status.creds_source is None:
            color = "white"
            text = "&lt;Refreshing&gt;"
        elif self._status.creds_source == api.AwsCredentialsSource.NOT_VALID:
            color = "red"
            text = self._status.creds_source.name
        else:
            color = "green"
            text = self._status.creds_source.name
        self.creds_source_group.label.setText(f"<b style='color:{color};'>{text}</b>")

    def _auth_status_changed(self) -> None:
        if self._status.auth_status is None:
            color = "white"
            text = "&lt;Refreshing&gt;"
        elif self._status.auth_status == api.AwsAuthenticationStatus.AUTHENTICATED:
            color = "green"
            text = self._status.auth_status.name
        else:
            color = "red"
            text = self._status.auth_status.name
        self.auth_status_group.label.setText(f"<b style='color:{color};'>{text}</b>")

    def _api_availability_changed(self) -> None:
        if self._status.api_availability is None:
            color = "white"
            text = "&lt;Refreshing&gt;"
        elif self._status.api_availability:
            color = "green"
            text = "AUTHORIZED"
        else:
            color = "red"
            text = "UNAVAILABLE"
        self.deadline_authorized_group.label.setText(f"<b style='color:{color};'>{text}</b>")


class AuthenticationStatusGroup(QGroupBox):
    """
    UI element to group the status of authentication.
    """

    def __init__(self, *, title: str, parent: Optional[QWidget] = None):
        super().__init__(parent=parent, title=title)

        self._build_ui()

    def _build_ui(self):
        self.layout = QFormLayout(self)
        self.layout.setContentsMargins(0, 0, 0, 0)
        self.layout.setFieldGrowthPolicy(QFormLayout.AllNonFixedFieldsGrow)
        self.label = QLabel("")

        self.label.setAlignment(Qt.AlignCenter)
        self.layout.addWidget(self.label)
